"""
ANTsPy Registration
"""
__all__ = ["registration",
           "motion_correction",
           "label_image_registration"]

import numpy as np
from tempfile import mktemp
import glob
import re
import pandas as pd
import itertools
import os

import ants
from ants.internal import get_lib_fn, get_pointer_string, process_arguments

def registration(
    fixed,
    moving,
    type_of_transform="SyN",
    initial_transform=None,
    outprefix="",
    mask=None,
    moving_mask=None,
    mask_all_stages=False,
    grad_step=0.2,
    flow_sigma=3,
    total_sigma=0,
    aff_metric="mattes",
    aff_sampling=32,
    aff_random_sampling_rate=0.2,
    syn_metric="mattes",
    syn_sampling=32,
    reg_iterations=(40, 20, 0),
    aff_iterations=(2100, 1200, 1200, 10),
    aff_shrink_factors=(6, 4, 2, 1),
    aff_smoothing_sigmas=(3, 2, 1, 0),
    write_composite_transform=False,
    random_seed=None,
    verbose=False,
    multivariate_extras=None,
    restrict_transformation=None,
    smoothing_in_mm=False,
    singleprecision=True,
    use_legacy_histogram_matching=False,
    **kwargs
):
    """
    Register a pair of images either through the full or simplified
    interface to the ANTs registration method.

    ANTsR function: `antsRegistration`

    Arguments
    ---------
    fixed : ANTsImage
        fixed image to which we register the moving image.

    moving : ANTsImage
        moving image to be mapped to fixed space.

    type_of_transform : string
        A linear or non-linear registration type. Mutual information metric by default.
        See Notes below for more.

    initial_transform : list of strings (optional)
        transforms to prepend. If None, a translation is computed to align the image centers of mass, unless the type of
        transform is deformable-only (time-varying diffeomorphisms, SyNOnly, or antsRegistrationSyN*[so|bo]).
        To force initialization with an identity transform, set this to 'Identity'.

    outprefix : string
        output will be named with this prefix.

    mask : ANTsImage (optional)
        Registration metric mask in the fixed image space.

    moving_mask : ANTsImage (optional)
        Registration metric mask in the moving image space.

    mask_all_stages : boolean
        If true, apply metric mask(s) to all registration stages, instead of just the final stage.

    grad_step : scalar
        gradient step size (not for all tx)

    flow_sigma : scalar
        smoothing for update field
        At each iteration, the similarity metric and gradient is calculated.
        That gradient field is also called the update field and is smoothed
        before composing with the total field (i.e., the estimate of the total
        transform at that iteration). This total field can also be smoothed
        after each iteration.

    total_sigma : scalar
        smoothing for total field

    aff_metric : string
        the metric for the affine part (GC, mattes, meansquares)

    aff_sampling : scalar
        number of bins for the mutual information metric

    aff_random_sampling_rate : scalar
        the fraction of points used to estimate the metric. this can impact
        speed but also reproducibility and/or accuracy.

    syn_metric : string
        the metric for the syn part (CC, mattes, meansquares, demons)

    syn_sampling : scalar
        the nbins or radius parameter for the syn metric

    reg_iterations : list/tuple of integers
        vector of iterations for syn. we will set the smoothing and multi-resolution parameters based on the length of this vector.

    aff_iterations : list/tuple of integers
        vector of iterations for low-dimensional (translation, rigid, affine) registration.

    aff_shrink_factors : list/tuple of integers
        vector of multi-resolution shrink factors for low-dimensional (translation, rigid, affine) registration.

    aff_smoothing_sigmas : list/tuple of integers
        vector of multi-resolution smoothing factors for low-dimensional (translation, rigid, affine) registration.

    random_seed : integer
        random seed to improve reproducibility. note that the number of ITK_GLOBAL_DEFAULT_NUMBER_OF_THREADS should be 1 if you want perfect reproducibility.

    write_composite_transform : boolean
        Boolean specifying whether or not the composite transform (and its inverse, if it exists) should be written to an hdf5 composite file. This is false by default so that only the transform for each stage is written to file.

    verbose : boolean
        request verbose output (useful for debugging)

    multivariate_extras : additional metrics for multi-metric registration
        list of additional images and metrics which will
        trigger the use of multiple metrics in the registration
        process in the deformable stage. Each multivariate metric needs 5
        entries: name of metric, fixed, moving, weight,
        samplingParam. the list of lists should be of the form ( (
        "nameOfMetric2", img, img, weight, metricParam ) ). Another
        example would be  ( ( "MeanSquares", f2, m2, 0.5, 0
          ), ( "CC", f2, m2, 0.5, 2 ) ) .  This is only compatible
        with the SyNOnly or antsRegistrationSyN* transformations.

    restrict_transformation : This option allows the user to restrict the
          optimization of the displacement field, translation, rigid or
          affine transform on a per-component basis. For example, if
          one wants to limit the deformation or rotation of 3-D volume
          to the first two dimensions, this is possible by specifying a
          weight vector of ‘(1,1,0)’ for a 3D deformation field or
          ‘(1,1,0,1,1,0)’ for a rigid transformation. Restriction
          currently only works if there are no preceding
          transformations.

    smoothing_in_mm : boolean ; currently only impacts low dimensional registration

    singleprecision : boolean
        if True, use float32 for computations. This is useful for reducing memory
        usage for large datasets, at the cost of precision.

    use_legacy_histogram_matching : boolean
        if True, use the original histogram matching in ANTs. This is not recommended, but is available for backwards
        compatibilty with earlier versions, where it was always turned on. The default is False. A better implementation of
        histogram matching is available in the ants.histogram_match_image2 function.

    kwargs : keyword args
        extra arguments

    Returns
    -------
    dict containing follow key/value pairs:
        `warpedmovout`: Moving image warped to space of fixed image.
        `warpedfixout`: Fixed image warped to space of moving image.
        `fwdtransforms`: Transforms to move from moving to fixed image.
        `invtransforms`: Transforms to move from fixed to moving image.

    Notes
    -----
    type_of_transform can be one of:
        - "Translation": Translation transformation.
        - "Rigid": Rigid transformation: Only rotation and translation.
        - "Similarity": Similarity transformation: uniform scaling, rotation and translation.
        - "QuickRigid": Rigid transformation: Only rotation and translation.
                        May be useful for quick visualization fixes.'
        - "DenseRigid": Rigid transformation: Only rotation and translation.
                        Employs dense sampling during metric estimation.'
        - "BOLDRigid": Rigid transformation: Parameters typical for BOLD to
                        BOLD intrasubject registration'.'
        - "Affine": Affine transformation: Rigid + scaling + shear (12 parameters).
        - "AffineFast": Fast version of Affine.
        - "BOLDAffine": Affine transformation: Parameters typical for BOLD to
                        BOLD intrasubject registration'.'
        - "TRSAA": translation, rigid, similarity, affine (twice). please set
                    regIterations if using this option. this would be used in
                    cases where you want a really high quality affine mapping
                    (perhaps with mask).
        - "Elastic": Elastic deformation: Affine + deformable.
        - "ElasticSyN": Symmetric normalization: Affine + deformable
                        transformation, with mutual information as optimization
                        metric and elastic regularization.
        - "SyN": Symmetric normalization: Affine + deformable transformation,
                    with mutual information as optimization metric.
        - "SyNRA": Symmetric normalization: Rigid + Affine + deformable
                    transformation, with mutual information as optimization metric.
        - "SyNOnly": Symmetric normalization with no rigid or affine stages.
                    Uses mutual information as optimization metric.
        - "SyNCC": SyN, but with cross-correlation as the metric.
        - "SyNabp": SyN optimized for abpBrainExtraction.
        - "SyNBold": SyN, but optimized for registrations between BOLD and T1 images.
        - "SyNBoldAff": SyN, but optimized for registrations between BOLD
                        and T1 images, with additional affine step.
        - "SyNAggro": SyN, but with more aggressive registration
                        (fine-scale matching and more deformation).
                        Takes more time than SyN.
        - "SyNLessAggro": Does exactly the same thing as "SyNAggro".
        - "TV[n]": time-varying diffeomorphism with where 'n' indicates number of
            time points in velocity field discretization.  The initial transform
            should be computed, if needed, in a separate call to ants.registration.
        - "TVMSQ": time-varying diffeomorphism with mean square metric
        - "TVMSQC": time-varying diffeomorphism with mean square metric for very large deformation
        - "antsRegistrationSyN[x]": recreation of the antsRegistrationSyN.sh script in ANTs
                                    where 'x' is one of the transforms available:
                                         t: translation (1 stage)
                                         r: rigid (1 stage)
                                         a: rigid + affine (2 stages)
                                         s: rigid + affine + deformable syn (3 stages)
                                         sr: rigid + deformable syn (2 stages)
                                         so: deformable syn only (1 stage)
                                         b: rigid + affine + deformable b-spline syn (3 stages)
                                         br: rigid + deformable b-spline syn (2 stages)
                                         bo: deformable b-spline syn only (1 stage)
        - "antsRegistrationSyNQuick[x]": recreation of the antsRegistrationSyNQuick.sh script in ANTs.
                                         x options as above.
        - "antsRegistrationSyNRepro[x]": reproducible registration.  x options as above.
        - "antsRegistrationSyNQuickRepro[x]": quick reproducible registration.  x options as above.

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('r16'))
    >>> mi = ants.image_read(ants.get_ants_data('r64'))
    >>> fi = ants.resample_image(fi, (60,60), 1, 0)
    >>> mi = ants.resample_image(mi, (60,60), 1, 0)
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'SyN' )
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[t]' )
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[b]' )
    >>> mytx = ants.registration(fixed=fi, moving=mi, type_of_transform = 'antsRegistrationSyN[s]' )
    """
    if isinstance(fixed, list) and (moving is None):
        processed_args = process_arguments(fixed)
        libfn = get_lib_fn("antsRegistration")
        reg_exit = libfn(processed_args)
        if (reg_exit != 0):
            raise RuntimeError(f"Registration failed with error code {reg_exit}")
        else:
            return 0

    if not (ants.is_image(fixed) and ants.is_image(moving)):
        raise ValueError("Fixed and moving images must be ANTsImage objects")

    if type_of_transform == "":
        type_of_transform = "SyN"

    if isinstance(type_of_transform, (tuple, list)) and (len(type_of_transform) == 1):
        type_of_transform = type_of_transform[0]

    if (outprefix == "") or len(outprefix) == 0:
        outprefix = mktemp()

    if np.sum(np.isnan(fixed.numpy())) > 0:
        raise ValueError("fixed image has NaNs - replace these")
    if np.sum(np.isnan(moving.numpy())) > 0:
        raise ValueError("moving image has NaNs - replace these")

    if fixed.dimension != moving.dimension:
        raise ValueError("Fixed and moving image dimensions are not the same.")
    # ----------------------------

    myiterations = aff_iterations
    args = [fixed, moving, type_of_transform, outprefix]
    shrinkfactors_affine = "6x4x2x1"  # old fixed params
    smoothingsigmas_affine = "3x2x1x0"  # old fixed params
    if (
        type(aff_shrink_factors) is int
        or type(aff_smoothing_sigmas) is int
        or type(aff_iterations) is int
    ):
        if type(aff_smoothing_sigmas) is not int:
            raise ValueError("aff_smoothing_sigmas should be a single integer.")
        if type(aff_iterations) is not int:
            raise ValueError("aff_iterations should be a single integer.")
        if type(aff_shrink_factors) is not int:
            raise ValueError("aff_shrink_factors should be a single integer.")
        shrinkfactors_affine = aff_shrink_factors
        smoothingsigmas_affine = aff_smoothing_sigmas
        myiterations = aff_iterations

    if restrict_transformation is not None:
        if type(restrict_transformation) is tuple:
            restrict_transformationchar = "x".join([str(ri) for ri in restrict_transformation])

    if type(aff_shrink_factors) is tuple:
        shrinkfactors_affine = "x".join([str(ri) for ri in aff_shrink_factors])
        smoothingsigmas_affine = "x".join([str(ri) for ri in aff_smoothing_sigmas])
        myiterations = "x".join([str(ri) for ri in aff_iterations])
        if len(aff_iterations) != len(aff_smoothing_sigmas):
            raise ValueError(
                "aff_iterations length should equal aff_smoothing_sigmas length."
            )
        if len(aff_iterations) != len(aff_shrink_factors):
            raise ValueError(
                "aff_iterations length should equal aff_shrink_factors length."
            )
        if len(aff_shrink_factors) != len(aff_smoothing_sigmas):
            raise ValueError(
                "aff_shrink_factors length should equal aff_smoothing_sigmas length."
            )

    if type_of_transform == "AffineFast":
        type_of_transform = "Affine"
        myiterations = "2100x1200x0x0"
    if type_of_transform == "BOLDAffine":
        type_of_transform = "Affine"
        shrinkfactors_affine = "2x1"
        smoothingsigmas_affine = "1x0"
        myiterations = "100x20"
    if type_of_transform == "QuickRigid":
        type_of_transform = "Rigid"
        myiterations = "20x20x0x0"
    if type_of_transform == "DenseRigid":
        type_of_transform = "Rigid"
        aff_random_sampling_rate = 1.0
    if type_of_transform == "BOLDRigid":
        type_of_transform = "Rigid"
        shrinkfactors_affine = "2x1"
        smoothingsigmas_affine = "1x0"
        myiterations = "100x20"

    if smoothing_in_mm:
        smoothingsigmas_affine = smoothingsigmas_affine + 'mm'

    mysyn = "SyN[%f,%f,%f]" % (grad_step, flow_sigma, total_sigma)
    if type_of_transform == "Elastic":
        mysyn = "GaussianDisplacementField[%f,%f,%f]" % (grad_step, flow_sigma, total_sigma)
    itlen = len(reg_iterations)  # NEED TO CHECK THIS
    if itlen == 0:
        smoothingsigmas = 0
        shrinkfactors = 1
        synits = reg_iterations
    else:
        smoothingsigmas = np.arange(0, itlen)[::-1].astype(
            "float32"
        )  # NEED TO CHECK THIS
        shrinkfactors = 2 ** smoothingsigmas
        shrinkfactors = shrinkfactors.astype("int")
        smoothingsigmas = "x".join([str(ss)[0] for ss in smoothingsigmas])
        shrinkfactors = "x".join([str(ss) for ss in shrinkfactors])
        synits = "x".join([str(ri) for ri in reg_iterations])

    inpixeltype = fixed.pixeltype
    output_pixel_type = 'float' if singleprecision else 'double'

    tvTypes = [
        "TV[1]",
        "TV[2]",
        "TV[3]",
        "TV[4]",
        "TV[5]",
        "TV[6]",
        "TV[7]",
        "TV[8]",
    ]
    allowable_tx = {
        "SyNBold",
        "SyNBoldAff",
        "ElasticSyN",
        "Elastic",
        "SyN",
        "SyNRA",
        "SyNOnly",
        "SyNAggro",
        "SyNCC",
        "TRSAA",
        "SyNabp",
        "SyNLessAggro",
        "TV[1]",
        "TV[2]",
        "TV[3]",
        "TV[4]",
        "TV[5]",
        "TV[6]",
        "TV[7]",
        "TV[8]",
        "TVMSQ",
        "TVMSQC",
        "Rigid",
        "Similarity",
        "Translation",
        "Affine",
        "AffineFast",
        "BOLDAffine",
        "QuickRigid",
        "DenseRigid",
        "BOLDRigid"
    }
    deformable_only_transforms = [
        "SyNOnly",
        "antsRegistrationSyN[so]",
        "antsRegistrationSyNQuick[so]",
        "antsRegistrationSyNRepro[so]",
        "antsRegistrationSyNQuickRepro[so]",
        "antsRegistrationSyN[bo]",
        "antsRegistrationSyNQuick[bo]",
        "antsRegistrationSyNRepro[bo]",
        "antsRegistrationSyNQuickRepro[bo]",
        "TVMSQ",
        "TVMSQC"
    ] + tvTypes
    transform_type_exists = type_of_transform in allowable_tx

    # Perform checking of antsRegistrationSyN transforms later
    if not "antsRegistrationSyN" in type_of_transform and not transform_type_exists:
        raise ValueError(f'{type_of_transform} does not exist')

    if isinstance(initial_transform, str):
        initial_transform = [initial_transform]
    # if isinstance(initx, ANTsTransform):
    # tempTXfilename = tempfile( fileext = '.mat' )
    # initx = invertAntsrTransform( initialTransform )
    # initx = invertAntsrTransform( initx )
    # writeAntsrTransform( initx, tempTXfilename )
    # initial_transform = tempTXfilename
    moving = moving.clone(output_pixel_type)
    fixed = fixed.clone(output_pixel_type)
    # NOTE: this may be better for general purpose applications: TBD
#    moving = ants.iMath( moving.clone("float"), "Normalize" )
#    fixed = ants.iMath( fixed.clone("float"), "Normalize" )
    warpedfixout = moving.clone()
    warpedmovout = fixed.clone()
    fixed_str = get_pointer_string(fixed)
    moving_str = get_pointer_string(moving)
    warpedfixout_str = get_pointer_string(warpedfixout)
    warpedmovout_str = get_pointer_string(warpedmovout)
    if mask is not None:
        mask_binary = mask != 0
        fixed_mask_str = get_pointer_string(mask_binary)
    else:
        fixed_mask_str = "NA"

    if moving_mask is not None:
        moving_mask_binary = moving_mask != 0
        moving_mask_str = get_pointer_string(moving_mask_binary)
    else:
        moving_mask_str = "NA"

    maskopt = "[%s,%s]" % (fixed_mask_str, moving_mask_str)

    if mask_all_stages:
        earlymaskopt = maskopt
    else:
        earlymaskopt = "[NA,NA]"

    if initial_transform is None:
        if type_of_transform in deformable_only_transforms:
            initial_transform = ["Identity"]
        else:
            initial_transform = ["[%s,%s,1]" % (fixed_str, moving_str)]

    # ------------------------------------------------------------
    if type_of_transform == "SyNBold":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Rigid[0.25]",
            "--convergence", "[1200x1200x100,1e-6,5]",
            "--smoothing-sigmas", "2x1x0",
            "--shrink-factors", "4x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyNBoldAff":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Rigid[0.25]",
            "--convergence", "[1200x1200x100,1e-6,5]",
            "--smoothing-sigmas", "2x1x0",
            "--shrink-factors", "4x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "[200x20,1e-6,5]",
            "--smoothing-sigmas", "1x0",
            "--shrink-factors", "2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % (synits),
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "ElasticSyN":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "2100x1200x200x0",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % (synits),
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyN" or type_of_transform == "Elastic":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "2100x1200x1200x0",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyNRA":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Rigid[0.25]",
            "--convergence", "2100x1200x1200x0",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "2100x1200x1200x0",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyNOnly":
        if multivariate_extras is None:
            args = [
                "--dimensionality", str(fixed.dimension),
                "-r"] + initial_transform + [
                "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
                "--transform", mysyn,
                "--convergence", "[%s,1e-7,8]" % synits,
                "--smoothing-sigmas", smoothingsigmas,
                "--shrink-factors", shrinkfactors,
                "-u", str(int(use_legacy_histogram_matching)),
                "-z", "1",
                "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            ]
        else:
            metrics = []
            for mve_idx in range(len(multivariate_extras)):
                metrics.append("--metric")
                metric_name = multivariate_extras[mve_idx][0]
                metric_fixed_str = get_pointer_string(
                    multivariate_extras[mve_idx][1]
                )
                metric_moving_str = get_pointer_string(
                    multivariate_extras[mve_idx][2]
                )
                metric_weight = multivariate_extras[mve_idx][3]
                metric_sampling = multivariate_extras[mve_idx][4]
                metric_full_string = "%s[%s,%s,%s,%s]" % (
                    metric_name,
                    metric_fixed_str,
                    metric_moving_str,
                    metric_weight,
                    metric_sampling,
                )
                metrics.append(metric_full_string)
            args_pre = [
                "--dimensionality", str(fixed.dimension),
                "-r"] + initial_transform + [
                "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            ]
            args_post = [
                "--transform", mysyn,
                "--convergence", "[%s,1e-7,8]" % synits,
                "--smoothing-sigmas", smoothingsigmas,
                "--shrink-factors", shrinkfactors,
                "-u", str(int(use_legacy_histogram_matching)),
                "-z", "1",
                "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            ]
            args = args_pre + metrics + args_post
        args.append("-x")
        args.append(maskopt)
    # ------------------------------------------------------------
    elif type_of_transform == "SyNAggro":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "2100x1200x1200x100",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyNCC":
        # syn_metric = "CC"
        # syn_sampling = 4
        # synits = "2100x1200x1200x20"
        # smoothingsigmas = "3x2x1x0"
        # shrinkfactors = "4x3x2x1"
        # mysyn = "SyN[0.15,3,0]"
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Rigid[1]",
            "--convergence", "2100x1200x1200x0",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x4x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[1]",
            "--convergence", "1200x1200x100",
            "--smoothing-sigmas", "2x1x0",
            "--shrink-factors", "4x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % ("CC", fixed_str, moving_str, 4),
            "--transform", "SyN[0.15,3,0]",
            "--convergence", "[2100x1200x1200x20,1e-7,8]",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x3x2x1",
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "TRSAA":
        itlen = len(reg_iterations)
        itlenlow = round(itlen / 2 + 0.0001)
        dlen = itlen - itlenlow
        _myconvlow = [2000] * itlenlow + [0] * dlen
        myconvlow = "x".join([str(mc) for mc in _myconvlow])
        myconvhi = "x".join([str(r) for r in reg_iterations])
        myconvhi = "[%s,1.e-7,10]" % myconvhi
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Translation[1]",
            "--convergence", myconvlow,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Rigid[1]",
            "--convergence", myconvlow,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Similarity[1]",
            "--convergence", myconvlow,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[1]",
            "--convergence", myconvhi,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[1]",
            "--convergence", myconvhi,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------s
    elif type_of_transform == "SyNabp":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "mattes[%s,%s,1,32,regular,0.25]" % (fixed_str, moving_str),
            "--transform", "Rigid[0.1]",
            "--convergence", "1000x500x250x100",
            "--smoothing-sigmas", "4x2x1x0",
            "--shrink-factors", "8x4x2x1",
            "-x", earlymaskopt,
            "--metric", "mattes[%s,%s,1,32,regular,0.25]" % (fixed_str, moving_str),
            "--transform", "Affine[0.1]",
            "--convergence", "1000x500x250x100",
            "--smoothing-sigmas", "4x2x1x0",
            "--shrink-factors", "8x4x2x1",
            "-x", earlymaskopt,
            "--metric", "CC[%s,%s,0.5,4]" % (fixed_str, moving_str),
            "--transform", "SyN[0.1,3,0]",
            "--convergence", "50x10x0",
            "--smoothing-sigmas", "2x1x0",
            "--shrink-factors", "4x2x1",
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "SyNLessAggro":
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "Affine[0.25]",
            "--convergence", "2100x1200x1200x100",
            "--smoothing-sigmas", "3x2x1x0",
            "--shrink-factors", "4x2x2x1",
            "-x", earlymaskopt,
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", mysyn,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform in tvTypes:
        if grad_step is None:
            grad_step = 1.0
        nTimePoints = type_of_transform.split("[")[1].split("]")[0]
        tvtx = (
            "TimeVaryingVelocityField["
            + ",".join([
                str(grad_step),
                nTimePoints,
                str(flow_sigma),
                "0.0",
                str(total_sigma),
                "0"])
            +"]"
        )
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", tvtx,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "0",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    elif type_of_transform == "TVMSQ":
        if grad_step is None:
            grad_step = 1.0
        tvtx = "TimeVaryingVelocityField[%s, 4, 0.0,0.0, 0.5,0 ]" % str(
            grad_step
        )
        args = [
            "--dimensionality",
            str(fixed.dimension),
            '-r'
        ] + initial_transform + [
            "--metric",
            "%s[%s,%s,1,%s]" % (syn_metric, fixed_str, moving_str, syn_sampling),
            "--transform", tvtx,
            "--convergence", "[%s,1e-7,8]" % synits,
            "--smoothing-sigmas", smoothingsigmas,
            "--shrink-factors", shrinkfactors,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "0",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform == "TVMSQC":
        if grad_step is None:
            grad_step = 2.0

        tvtx = "TimeVaryingVelocityField[%s, 8, 1.0,0.0, 0.05,0 ]" % str(
            grad_step
        )
        args = [
            "--dimensionality", str(fixed.dimension),
            '-r'] + initial_transform + [
            "--metric", "demons[%s,%s,0.5,0]" % (fixed_str, moving_str),
            "--metric", "meansquares[%s,%s,1,0]" % (fixed_str, moving_str),
            "--transform", tvtx,
            "--convergence", "[1200x1200x100x20x0,0,5]",
            "--smoothing-sigmas", "8x6x4x2x1vox",
            "--shrink-factors", "8x6x4x2x1",
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "0",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif type_of_transform in ("Rigid", "Similarity", "Translation", "Affine"):
        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--metric", "%s[%s,%s,1,%s,regular,%s]" % (aff_metric, fixed_str, moving_str, aff_sampling, aff_random_sampling_rate),
            "--transform", "%s[0.25]" % type_of_transform,
            "--convergence", myiterations,
            "--smoothing-sigmas", smoothingsigmas_affine,
            "--shrink-factors", shrinkfactors_affine,
            "-u", str(int(use_legacy_histogram_matching)),
            "-z", "1",
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
            "-x", maskopt
        ]
    # ------------------------------------------------------------
    elif "antsRegistrationSyN" in type_of_transform:
        do_quick = ("Quick" in type_of_transform)
        do_repro = ("Repro" in type_of_transform)

        subtype_of_transform = "s"
        spline_distance = 26
        metric_parameter = 32 if do_quick else 2
        linear_metric = "GC[%s,%s,1,1,Regular,0.25]" if do_repro else "MI[%s,%s,1,32,Regular,0.25]"

        rigid_shrink_factors = "8x4x2x1"
        rigid_smoothing_sigmas = "3x2x1x0vox"

        affine_shrink_factors = "8x4x2x1"
        affine_smoothing_sigmas = "3x2x1x0vox"
        
        linear_gradient_step = 0.1
        syn_gradient_step = 0.2

        if "[" in type_of_transform and "]" in type_of_transform:
            subtype_of_transform = type_of_transform.split("[")[1].split(
                "]"
            )[0]
            if "," in subtype_of_transform:
                subtype_of_transform_args = subtype_of_transform.split(",")
                subtype_of_transform = subtype_of_transform_args[0]
                if not ( subtype_of_transform in ["b", "br", "bo", "s", "sr", "so"]):
                    raise ValueError("Extra parameters are only valid for 's' or 'b' SyN transforms.")
                metric_parameter = subtype_of_transform_args[1]
                if len(subtype_of_transform_args) > 2:
                    spline_distance = subtype_of_transform_args[2]

        if do_quick:
            rigid_convergence = "[1000x500x250x0,1e-6,10]"
            affine_convergence = "[1000x500x250x0,1e-6,10]"
            syn_convergence = "[100x70x50x0,1e-6,10]"
            if do_repro:
                metric_parameter = 2
                syn_metric = "CC[%s,%s,1,%s]" % (fixed_str, moving_str, metric_parameter)
            else:
                metric_parameter = 32
                syn_metric = "MI[%s,%s,1,%s]" % (fixed_str, moving_str, metric_parameter)
        else:
            rigid_convergence = "[1000x500x250x100,1e-6,10]"
            affine_convergence = "[1000x500x250x100,1e-6,10]"
            syn_convergence = "[100x70x50x20,1e-6,10]"
            metric_parameter = 2
            syn_metric = "CC[%s,%s,1,%s]" % (fixed_str, moving_str, metric_parameter)
        syn_shrink_factors = "8x4x2x1"
        syn_smoothing_sigmas = "3x2x1x0vox"
        if subtype_of_transform in ("sr", "br"):
            if do_quick:
                syn_convergence = "[50x0,1e-6,10]"
            else:
                syn_convergence = "[50x20,1e-6,10]"
            syn_shrink_factors = "2x1"
            syn_smoothing_sigmas = "1x0vox"

        if random_seed is None and do_repro:
            random_seed = "1"

        rigidtx = "Translation" if subtype_of_transform == "t" else "Rigid"

        rigid_stage = [
            "--transform", rigidtx + "[" + str(linear_gradient_step) + "]",
            "--metric", linear_metric % (fixed_str, moving_str),
            "--convergence", rigid_convergence,
            "--shrink-factors", rigid_shrink_factors,
            "--smoothing-sigmas", rigid_smoothing_sigmas,
        ]
        affine_stage = [
            "--transform", "Affine[" + str(linear_gradient_step) + "]",
            "--metric", linear_metric % (fixed_str, moving_str),
            "--convergence", affine_convergence,
            "--shrink-factors", affine_shrink_factors,
            "--smoothing-sigmas", affine_smoothing_sigmas,
        ]
        syn_stage = [
            "--metric", syn_metric,
        ]
        if multivariate_extras is not None:
            for mve_idx in range(len(multivariate_extras)):
                syn_stage.append("--metric")
                metric_name = multivariate_extras[mve_idx][0]
                metric_fixed_str = get_pointer_string(
                    multivariate_extras[mve_idx][1]
                )
                metric_moving_str = get_pointer_string(
                    multivariate_extras[mve_idx][2]
                )
                metric_weight = multivariate_extras[mve_idx][3]
                metric_sampling = multivariate_extras[mve_idx][4]
                metric_full_string = "%s[%s,%s,%s,%s]" % (
                    metric_name,
                    metric_fixed_str,
                    metric_moving_str,
                    metric_weight,
                    metric_sampling,
                )
                syn_stage.append(metric_full_string)

        syn_stage.append("--convergence")
        syn_stage.append(syn_convergence)
        syn_stage.append("--shrink-factors")
        syn_stage.append(syn_shrink_factors)
        syn_stage.append("--smoothing-sigmas")
        syn_stage.append(syn_smoothing_sigmas)

        if subtype_of_transform in ("b", "br", "bo"):
            syn_stage.insert(0, "BSplineSyN[" + str(syn_gradient_step) + "," + str(spline_distance) + ",0,3]")
            syn_stage.insert(0, "--transform")

        if subtype_of_transform in ("s", "sr", "so"):
            syn_stage.insert(0, "SyN[" + str(syn_gradient_step) + ",3,0]")
            syn_stage.insert(0, "--transform")

        args = [
            "--dimensionality", str(fixed.dimension),
            "-r"] + initial_transform + [
            "--output", "[%s,%s,%s]" % (outprefix, warpedmovout_str, warpedfixout_str),
        ]

        if subtype_of_transform == "r" or subtype_of_transform == "t":
            args.append(rigid_stage)
        if subtype_of_transform == "a":
            args.append(rigid_stage)
            args.append(affine_stage)
        if subtype_of_transform == "b" or subtype_of_transform == "s":
            args.append(rigid_stage)
            args.append(affine_stage)
            args.append(syn_stage)
        if subtype_of_transform == "br" or subtype_of_transform == "sr":
            args.append(rigid_stage)
            args.append(syn_stage)
        if subtype_of_transform == "bo" or subtype_of_transform == "so":
            args.append(syn_stage)

        args.append("-x")
        args.append(maskopt)

        args = list(
            itertools.chain.from_iterable(
                itertools.repeat(x, 1) if isinstance(x, str) else x
                for x in args
            )
        )

    # ------------------------------------------------------------

    if random_seed is not None:
        args.append("--random-seed")
        args.append(random_seed)

    if restrict_transformation is not None:
        args.append("-g")
        args.append(restrict_transformationchar)

    args.append("--float")
    args.append(str(int(singleprecision)))
    args.append("--write-composite-transform")
    args.append(write_composite_transform * 1)
    if verbose:
        args.append("-v")
        args.append("1")

    processed_args = process_arguments(args)
    libfn = get_lib_fn("antsRegistration")
    if verbose:
        print("antsRegistration " + ' '.join(processed_args))
    reg_exit = libfn(processed_args)
    if (reg_exit != 0):
        raise RuntimeError(f"Registration failed with error code {reg_exit}")
    afffns = glob.glob(outprefix + "*" + "[0-9]GenericAffine.mat")
    fwarpfns = glob.glob(outprefix + "*" + "[0-9]Warp.nii.gz")
    iwarpfns = glob.glob(outprefix + "*" + "[0-9]InverseWarp.nii.gz")
    vfieldfns = glob.glob(outprefix + "*" + "[0-9]VelocityField.nii.gz")
    # print(afffns, fwarpfns, iwarpfns)
    if len(afffns) == 0:
        afffns = ""
    if len(fwarpfns) == 0:
        fwarpfns = ""
    if len(iwarpfns) == 0:
        iwarpfns = ""
    if len(vfieldfns) == 0:
        vfieldfns = ""

    alltx = sorted(
        set(glob.glob(outprefix + "*" + "[0-9]*"))
        - set(glob.glob(outprefix + "*VelocityField*"))
    )
    findinv = np.where(
        [re.search("[0-9]InverseWarp.nii.gz", ff) for ff in alltx]
    )[0]
    findfwd = np.where([re.search("[0-9]Warp.nii.gz", ff) for ff in alltx])[
        0
    ]
    if len(findinv) > 0:
        fwdtransforms = list(
            reversed(
                [ff for idx, ff in enumerate(alltx) if idx != findinv[0]]
            )
        )
        invtransforms = [
            ff for idx, ff in enumerate(alltx) if idx != findfwd[0]
        ]
    else:
        fwdtransforms = list(reversed(alltx))
        invtransforms = alltx

    if write_composite_transform:
        fwdtransforms = outprefix + "Composite.h5"
        invtransforms = outprefix + "InverseComposite.h5"

    if not vfieldfns:
        return {
            "warpedmovout": warpedmovout.clone(inpixeltype),
            "warpedfixout": warpedfixout.clone(inpixeltype),
            "fwdtransforms": fwdtransforms,
            "invtransforms": invtransforms,
        }
    else:
        return {
            "warpedmovout": warpedmovout.clone(inpixeltype),
            "warpedfixout": warpedfixout.clone(inpixeltype),
            "fwdtransforms": fwdtransforms,
            "invtransforms": invtransforms,
            "velocityfield": vfieldfns,
        }

def motion_correction(
    image,
    fixed=None,
    type_of_transform="BOLDRigid",
    mask=None,
    fdOffset=50,
    outprefix="",
    verbose=False,
    **kwargs
):
    """
    Correct time-series data for motion.

    ANTsR function: `antsrMotionCalculation`

    Arguments
    ---------
        image: antsImage, usually ND where D=4.

        fixed: Fixed image to register all timepoints to.  If not provided,
            mean image is used.

        type_of_transform : string
            A linear or non-linear registration type. Mutual information metric and rigid transformation by default.
            See ants registration for details.

        mask: mask for image (ND-1).  If not provided, estimated from data.
            2023-02-05: a performance change - previously, we estimated a mask
            when None is provided and would pass this to the registration.  this
            impairs performance if the mask estimate is bad.  in such a case, we
            prefer no mask at all.  As such, we no longer pass the mask to the
            registration when None is provided.

        fdOffset: offset value to use in framewise displacement calculation

        outprefix : string
            output will be named with this prefix plus a numeric extension.

        verbose: boolean

        kwargs: keyword args
            extra arguments - these extra arguments will control the details of registration that is performed. see ants registration for more.

    Returns
    -------
    dict containing follow key/value pairs:
        `motion_corrected`: Moving image warped to space of fixed image.
        `motion_parameters`: transforms for each image in the time series.
        `FD`: Framewise displacement generalized for arbitrary transformations.

    Notes
    -----
    Control extra arguments via kwargs. see ants.registration for details.

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read(ants.get_ants_data('ch2'))
    >>> mytx = ants.motion_correction( fi )
    """
    idim = image.dimension
    ishape = image.shape
    nTimePoints = ishape[idim - 1]
    if fixed is None:
        wt = 1.0 / nTimePoints
        fixed = ants.slice_image(image, axis=idim - 1, idx=0) * 0
        for k in range(nTimePoints):
            temp = ants.slice_image(image, axis=idim - 1, idx=k)
            fixed = fixed + ants.iMath(temp,"Normalize") * wt
    if mask is None:
        mask = ants.get_mask(fixed)
        useMask=None
    else:
        useMask=mask
    FD = np.zeros(nTimePoints)
    motion_parameters = list()
    motion_corrected = list()
    centerOfMass = mask.get_center_of_mass()
    npts = pow(2, idim - 1)
    pointOffsets = np.zeros((npts, idim - 1))
    myrad = np.ones(idim - 1).astype(int).tolist()
    mask1vals = np.zeros(int(mask.sum()))
    mask1vals[round(len(mask1vals) / 2)] = 1
    mask1 = ants.make_image(mask, mask1vals)
    myoffsets = ants.get_neighborhood_in_mask(
        mask1, mask1, radius=myrad, spatial_info=True
    )["offsets"]

    mycols = list("xy")
    if idim - 1 == 3:
        mycols = list("xyz")
    useinds = list()
    for k in range(myoffsets.shape[0]):
        if abs(myoffsets[k, :]).sum() == (idim - 2):
            useinds.append(k)
        myoffsets[k, :] = myoffsets[k, :] * fdOffset / 2.0 + centerOfMass
    fdpts = pd.DataFrame(data=myoffsets[useinds, :], columns=mycols)
    if verbose:
        print("Progress:")
    counter = 0
    for k in range(nTimePoints):
        mycount = round(k / nTimePoints * 100)
        if verbose and mycount == counter:
            counter = counter + 10
            print(mycount, end="%.", flush=True)
        temp = ants.slice_image(image, axis=idim - 1, idx=k)
        temp = ants.iMath(temp, "Normalize")
        if temp.numpy().var() > 0:
            if outprefix != "":
                outprefixloc = outprefix + "_" + str.zfill( str(k), 5 ) + "_"
                myreg = registration(
                    fixed, temp, type_of_transform=type_of_transform, mask=useMask,
                    outprefix=outprefixloc, **kwargs
                )
            else:
                myreg = registration(
                    fixed, temp, type_of_transform=type_of_transform, mask=useMask, **kwargs
                )
            fdptsTxI = ants.apply_transforms_to_points(
                idim - 1, fdpts, myreg["fwdtransforms"]
            )
            if k > 0 and motion_parameters[k - 1] != "NA":
                fdptsTxIminus1 = ants.apply_transforms_to_points(
                    idim - 1, fdpts, motion_parameters[k - 1]
                )
            else:
                fdptsTxIminus1 = fdptsTxI
            # take the absolute value, then the mean across columns, then the sum
            FD[k] = (fdptsTxIminus1 - fdptsTxI).abs().mean().sum()
            motion_parameters.append(myreg["fwdtransforms"])
            mywarped = ants.apply_transforms( fixed,
                ants.slice_image(image, axis=idim - 1, idx=k),
                myreg["fwdtransforms"] )
            motion_corrected.append(mywarped)
        else:
            motion_parameters.append("NA")
            motion_corrected.append(temp)

    if verbose:
        print("Done")
    return {
        "motion_corrected": ants.list_to_ndimage(image, motion_corrected),
        "motion_parameters": motion_parameters,
        "FD": FD,
    }

def label_image_registration(fixed_label_images,
                             moving_label_images,
                             fixed_intensity_images=None,
                             moving_intensity_images=None,
                             fixed_mask=None,
                             moving_mask=None,
                             initial_transforms='affine',
                             type_of_deformable_transform='antsRegistrationSyNQuick[so]',
                             label_image_weighting=1.0,
                             output_prefix='',
                             random_seed=None,
                             verbose=False):

    """
    Perform pairwise registration using fixed and moving sets of label
    images (and, optionally, sets of corresponding intensity images).

    Arguments
    ---------
    fixed_label_images : single or list of ANTsImage
        A single (or set of) fixed label image(s).

    moving_label_images : single or list of ANTsImage
        A single (or set of) moving label image(s).

    fixed_intensity_images : single or list of ANTsImage
        Optional---a single (or set of) fixed intensity image(s).

    moving_intensity_images : single or list of ANTsImage
        Optional---a single (or set of) moving intensity image(s).

    fixed_mask : ANTsImage
        Defines region for similarity metric calculation in the space
        of the fixed image.

    moving_mask : ANTsImage
        Defines region for similarity metric calculation in the space
        of the moving image.

    initial_transforms : string or list of files
        If specified, there are two options:  2) Use label images with 
        the centers of mass to a calculate linear transform of type 
        'identity', 'rigid', 'similarity', or 'affine'.  2) Specify a
        list of transform files, e.g., the output of ants.registration().

    type_of_deformable_transform : string
        Only works with deformable-only transforms, specifically the family
        of antsRegistrationSyN*[so] or antsRegistrationSyN*[bo] transforms.
        See 'type_of_transform' in ants.registration.  Additionally, one can
        use a list to pass a more tailored deformably-only transform
        optimization using SyN or BSplineSyN transforms.  The order of
        parameters in the list would be 1) transform specification, i.e.
        "SyN" or "BSplineSyN", 2) gradient (real), 3) intensity metric (string),
        4) intensity metric parameter (real), 5) convergence iterations per level
        (tuple) 6) smoothing factors per level (tuple), 7) shrink factors per level
        (tuple).  An example would type_of_deformable_transform = ["SyN", 0.2, "CC",
        4, (100,50,10), (2,1,0), (4,2,1)].

    label_image_weighting : float or list of floats
        Relative weighting for the label images.

    output_prefix : string
        Define the output prefix for the filenames of the output transform
        files.

    random_seed : integer
        Definition for deformable registration.

    verbose : boolean
        Print progress to the screen.

    Returns
    -------
    Set of transforms definining the mapping to/from the fixed image domain
    to the moving image domain.

    Example
    -------
    >>> import ants
    >>>
    >>> r16 = ants.image_read(ants.get_ants_data('r16'))
    >>> r16_seg1 = ants.threshold_image(r16, "Kmeans", 3) - 1
    >>> r16_seg2 = ants.threshold_image(r16, "Kmeans", 5) - 1
    >>> r64 = ants.image_read(ants.get_ants_data('r64'))
    >>> r64_seg1 = ants.threshold_image(r64, "Kmeans", 3) - 1
    >>> r64_seg2 = ants.threshold_image(r64, "Kmeans", 5) - 1
    >>> reg = ants.label_image_registration([r16_seg1, r16_seg2],
                                            [r64_seg1, r64_seg2],
                                            fixed_intensity_images=r16,
                                            moving_intensity_images=r64,
                                            initial_transforms='affine',
                                            type_of_deformable_transform='antsRegistrationSyNQuick[bo]',
                                            label_image_weighting=[1.0, 2.0],
                                            verbose=True)
    """

    # Perform validation check on the input

    if isinstance(fixed_label_images, ants.ANTsImage):
        fixed_label_images = [ants.image_clone(fixed_label_images)]
    if isinstance(moving_label_images, ants.ANTsImage):
        moving_label_images = [ants.image_clone(moving_label_images)]

    if len(fixed_label_images) != len(moving_label_images):
        raise ValueError("The number of fixed and moving label images do not match.")

    if fixed_intensity_images is not None or moving_intensity_images is not None:
        if isinstance(fixed_intensity_images, ants.ANTsImage):
            fixed_intensity_images = [ants.image_clone(fixed_intensity_images)]
        if isinstance(moving_intensity_images, ants.ANTsImage):
            moving_intensity_images = [ants.image_clone(moving_intensity_images)]
        if len(fixed_intensity_images) != len(moving_intensity_images):
            raise ValueError("The number of fixed and moving intensity images do not match.")

    label_image_weights = list()
    if isinstance(label_image_weighting, (int, float)):
        label_image_weights = [label_image_weighting] * len(fixed_label_images)
    else:
        label_image_weights = tuple(label_image_weighting)
        if len(fixed_label_images) != len(label_image_weights):
            raise ValueError("The length of label_image_weights must" +
                             "match the number of label image pairs.")

    image_dimension = fixed_label_images[0].dimension

    if output_prefix == "" or output_prefix is None or len(output_prefix) == 0:
        output_prefix = mktemp()

    do_deformable = True
    if type_of_deformable_transform is None or len(type_of_deformable_transform) == 0:
       do_deformable = False

    common_label_ids = list()
    total_number_of_labels = 0
    for i in range(len(fixed_label_images)):
        fixed_label_geoms = ants.label_geometry_measures(fixed_label_images[i])
        fixed_label_ids = np.array(fixed_label_geoms['Label'])
        moving_label_geoms = ants.label_geometry_measures(moving_label_images[i])
        moving_label_ids = np.array(moving_label_geoms['Label'])
        common_label_ids.append(np.intersect1d(moving_label_ids, fixed_label_ids))
        total_number_of_labels += len(common_label_ids[i])
        if verbose:
            print("Common label ids for image pair ", str(i), ": ", common_label_ids[i])
        if len(common_label_ids[i]) == 0:
            raise ValueError("No common labels for image pair " + str(i))

    deformable_multivariate_extras = list()

    if verbose:
        print("Total number of labels: " + str(total_number_of_labels))

    initial_xfrm_files = list()

    ##############################
    #
    #    Initial linear transform
    #
    ##############################

    if isinstance(initial_transforms, str) and initial_transforms in ['rigid', 'similarity', 'affine']:

        if verbose:
            print("\n\nComputing linear transform.\n")

        if total_number_of_labels < 3:
            raise ValueError("  Number of labels must be >= 3.")

        fixed_centers_of_mass = np.zeros((total_number_of_labels, image_dimension))
        moving_centers_of_mass = np.zeros((total_number_of_labels, image_dimension))
        count = 0
        for i in range(len(common_label_ids)):
            for j in range(len(common_label_ids[i])):
                label = common_label_ids[i][j]
                if verbose:
                    print("  Finding centers of mass for image pair " + str(i) + ", label " + str(label))
                fixed_single_label_image = ants.threshold_image(fixed_label_images[i], label, label, 1, 0)
                fixed_centers_of_mass[count, :] = ants.get_center_of_mass(fixed_single_label_image)
                moving_single_label_image = ants.threshold_image(moving_label_images[i], label, label, 1, 0)
                moving_centers_of_mass[count, :] = ants.get_center_of_mass(moving_single_label_image)
                count += 1
                if do_deformable:
                    deformable_multivariate_extras.append(["MSQ", fixed_single_label_image,
                                                           moving_single_label_image,
                                                           label_image_weights[i], 0])

        linear_xfrm = ants.fit_transform_to_paired_points(moving_centers_of_mass,
                                                          fixed_centers_of_mass,
                                                          transform_type=initial_transforms,
                                                          verbose=verbose)
        
        if do_deformable:
            linear_xfrm_file = output_prefix + "LandmarkBasedLinear" + initial_transforms + ".mat"
        else:
            linear_xfrm_file = output_prefix + "0GenericAffine.mat"

        ants.write_transform(linear_xfrm, linear_xfrm_file)
        initial_xfrm_files.append(linear_xfrm_file)

    elif initial_transforms is not None or initial_transforms == 'identity':
        
        if do_deformable:
            for i in range(len(common_label_ids)):
                for j in range(len(common_label_ids[i])):
                    label = common_label_ids[i][j]
                    fixed_single_label_image = ants.threshold_image(fixed_label_images[i], label, label, 1, 0)
                    moving_single_label_image = ants.threshold_image(moving_label_images[i], label, label, 1, 0)
                    deformable_multivariate_extras.append(["MSQ", fixed_single_label_image,
                                                        moving_single_label_image,
                                                        label_image_weights[i], 0])

        if initial_transforms != 'identity':
            if not isinstance(initial_transforms, list):
                initial_transforms = [initial_transforms]
            for i in range(len(initial_transforms)):
                if not os.path.exists(initial_transforms[i]):
                    raise ValueError(initial_transforms[i] + " does not exist.")
                else:
                    initial_xfrm_files.append(initial_transforms[i])
            
    ##############################
    #
    #    Deformable transform
    #
    ##############################

    if do_deformable:

        if verbose:
            print("\n\nComputing deformable transform using images.\n")

        intensity_metric = "CC"
        intensity_metric_parameter = 2
        syn_shrink_factors = "8x4x2x1"
        syn_smoothing_sigmas = "3x2x1x0vox"
        syn_convergence = "[100x70x50x20,1e-6,10]"
        spline_distance = 26
        gradient_step = 0.2
        syn_transform = "SyN"

        syn_stage = list()

        if isinstance(type_of_deformable_transform, list):

            if (len(type_of_deformable_transform) != 7 or
                not isinstance(type_of_deformable_transform[0], str) or
                not isinstance(type_of_deformable_transform[1], float) or
                not isinstance(type_of_deformable_transform[2], str) or
                not isinstance(type_of_deformable_transform[3], int) or
                not isinstance(type_of_deformable_transform[4], tuple) or
                not isinstance(type_of_deformable_transform[5], tuple) or
                not isinstance(type_of_deformable_transform[6], tuple)):
                raise ValueError("Incorrect specification for type_of_deformable_transform.  See help menu.")

            syn_transform = type_of_deformable_transform[0]
            gradient_step = type_of_deformable_transform[1]
            intensity_metric = type_of_deformable_transform[2]
            intensity_metric_parameter = type_of_deformable_transform[3]

            t = type_of_deformable_transform[4]
            tstr = ''.join(map(lambda x: str(x) + 'x', t[:len(t)-1])) + str(t[len(t)-1])
            syn_convergence = "[" + tstr + ",1e-6,10]"

            t = type_of_deformable_transform[5]
            tstr = ''.join(map(lambda x: str(x) + 'x', t[:len(t)-1])) + str(t[len(t)-1])
            syn_smoothing_sigmas = tstr + "vox"

            t = type_of_deformable_transform[6]
            syn_shrink_factors = ''.join(map(lambda x: str(x) + 'x', t[:len(t)-1])) + str(t[len(t)-1])

        else:

            do_quick = False
            if "Quick" in type_of_deformable_transform:
                do_quick = True
            elif "Repro" in type_of_deformable_transform:
                random_seed = str(1)

            if "[" in type_of_deformable_transform and "]" in type_of_deformable_transform:
                subtype_of_deformable_transform = type_of_deformable_transform.split("[")[1].split("]")[0]
                if not ('bo' in subtype_of_deformable_transform or 'so' in subtype_of_deformable_transform):
                    raise ValueError("Only 'so' or 'bo' transforms are available.")
                else:
                    if 'bo' in subtype_of_deformable_transform:
                        syn_transform = "BSplineSyN"
                if "," in subtype_of_deformable_transform:
                    subtype_of_deformable_transform_args = subtype_of_deformable_transform.split(",")
                    subtype_of_deformable_transform = subtype_of_deformable_transform_args[0]
                    intensity_metric_parameter = subtype_of_deformable_transform_args[1]
                    if len(subtype_of_deformable_transform_args) > 2:
                        spline_distance = subtype_of_deformable_transform_args[2]

            if do_quick:
                intensity_metric = "MI"
                intensity_metric_parameter = 32
                syn_convergence = "[100x70x50x0,1e-6,10]"

        if fixed_intensity_images is not None and len(fixed_intensity_images) > 0:
            for i in range(len(fixed_intensity_images)):
                syn_stage.append("--metric")
                metric_string = "%s[%s,%s,%s,%s]" % (
                    intensity_metric,
                    get_pointer_string(fixed_intensity_images[i]),
                    get_pointer_string(moving_intensity_images[i]),
                    1.0, intensity_metric_parameter)
                syn_stage.append(metric_string)

        for kk in range(len(deformable_multivariate_extras)):
            syn_stage.append("--metric")
            metricString = "%s[%s,%s,%s,%s]" % (
                "MSQ",
                get_pointer_string(deformable_multivariate_extras[kk][1]),
                get_pointer_string(deformable_multivariate_extras[kk][2]),
                deformable_multivariate_extras[kk][3], 0.0)
            syn_stage.append(metricString)

        syn_stage.append("--convergence")
        syn_stage.append(syn_convergence)
        syn_stage.append("--shrink-factors")
        syn_stage.append(syn_shrink_factors)
        syn_stage.append("--smoothing-sigmas")
        syn_stage.append(syn_smoothing_sigmas)

        if syn_transform == "SyN":
            syn_stage.insert(0, "SyN[" + str(gradient_step) + ",3,0]")
        else:
            syn_stage.insert(0, "BSplineSyN[" + str(gradient_step) + "," + str(spline_distance) + ",0,3]")
        syn_stage.insert(0, "--transform")

        args = ["--dimensionality", str(image_dimension),
                "--output", output_prefix]

        if len(initial_xfrm_files) > 0:
          for i in range(len(initial_xfrm_files)):
            initial_args = ["-r", initial_xfrm_files[i]]
            args.append(initial_args)

        args.append(syn_stage)

        fixed_mask_string = 'NA'
        if fixed_mask is not None:
            fixed_mask_binary = fixed_mask != 0
            fixed_mask_string = get_pointer_string(fixed_mask_binary)

        moving_mask_string = 'NA'
        if moving_mask is not None:
            moving_mask_binary = moving_mask != 0
            moving_mask_string = get_pointer_string(moving_mask_binary)

        mask_option = "[%s,%s]" % (fixed_mask_string, moving_mask_string)

        args.append("-x")
        args.append(mask_option)

        args = list(itertools.chain.from_iterable(
                    itertools.repeat(x, 1)
                    if isinstance(x, str)
                    else x for x in args))

        args.append("--float")
        args.append("1")

        if random_seed is not None:
            args.append("--random-seed")
            args.append(random_seed)

        if verbose:
            args.append("-v")
            args.append("1")

        processed_args = process_arguments(args)
        if verbose:
            print("antsRegistration " + ' '.join(processed_args))

        libfn = get_lib_fn("antsRegistration")
        deformable_registration_exit_error = libfn(processed_args)

        if deformable_registration_exit_error != 0:
            raise RuntimeError(f"Registration failed with error code {deformable_registration_exit_error}")

    all_xfrms = sorted(set(glob.glob(output_prefix + "*" + "[0-9]*")))

    find_inverse_warps_idx = np.where([re.search("[0-9]InverseWarp.nii.gz", ff) for ff in all_xfrms])[0]
    find_forward_warps_idx = np.where([re.search("[0-9]Warp.nii.gz", ff) for ff in all_xfrms])[0]
    find_affines_idx = np.where([re.search("[0-9]GenericAffine.mat", ff) for ff in all_xfrms])[0]

    fwdtransforms = list()
    invtransforms = list()

    if len(find_forward_warps_idx) > 0:
        fwdtransforms.append(all_xfrms[find_forward_warps_idx[0]])
    if len(find_affines_idx) > 0:    
        fwdtransforms.append(all_xfrms[find_affines_idx[0]])
        invtransforms.append(all_xfrms[find_affines_idx[0]])
    if len(find_inverse_warps_idx) > 0:
        invtransforms.append(all_xfrms[find_inverse_warps_idx[0]])

    if verbose:
        print("\n\nResulting transforms")
        print("  fwdtransforms: ", fwdtransforms)
        print("  invtransforms: ", invtransforms)

    return {
        "fwdtransforms": fwdtransforms,
        "invtransforms": invtransforms,
    }





__all__ = ["simulate_displacement_field"]

import numpy as np


import ants
from ants.internal import get_lib_fn



def simulate_displacement_field(domain_image, 
                                field_type="bspline", 
                                number_of_random_points=1000, 
                                sd_noise=10.0,
                                enforce_stationary_boundary=True,
                                number_of_fitting_levels=4,
                                mesh_size=1,
                                sd_smoothing=4.0):
    """
    simulate displacement field using either b-spline or exponential transform

    ANTsR function: `simulateDisplacementField`

    Arguments
    ---------
    domain_image : ANTsImage
        Domain image

    field_type : string
        Either "bspline" or "exponential".

    number_of_random_points : integer
        Number of displacement points.

    sd_noise : float
        Standard deviation of the displacement field noise.

    enforce_stationary_boundary : boolean
        Determines fixed boundary conditions.

    number_of_fitting_levels : integer
        Number of fitting levels (b-spline only).

    mesh_size : integer or n-D tuple
        Determines fitting resolution at base level (b-spline only).

    sd_smoothing : float
        Standard deviation of the Gaussian smoothing in mm (exponential only).
        
    Returns
    -------
    ANTs vector image.

    Example
    -------
    >>> import ants
    >>> domain = ants.image_read( ants.get_ants_data('r16'))
    >>> exp_field = ants.simulate_displacement_field(domain, field_type="exponential")
    >>> bsp_field = ants.simulate_displacement_field(domain, field_type="bspline")
    >>> bsp_xfrm = ants.transform_from_displacement_field(bsp_field * 3)
    >>> domain_warped = ants.apply_ants_transform_to_image(bsp_xfrm, domain, domain)
    """

    image_dimension = domain_image.dimension

    if field_type == 'bspline':
        if isinstance(mesh_size, int) == False and len(mesh_size) != image_dimension:
            raise ValueError("Incorrect specification for mesh_size.")

        spline_order = 3
        number_of_control_points = mesh_size + spline_order

        if isinstance(number_of_control_points, int) == True:
            number_of_control_points = np.repeat(number_of_control_points, image_dimension)

        libfn = get_lib_fn("simulateBsplineDisplacementField%iD" % image_dimension)
        field = libfn(domain_image.pointer, number_of_random_points, sd_noise, 
                      enforce_stationary_boundary, number_of_fitting_levels, number_of_control_points)
        bspline_field = ants.from_pointer(field).clone('float')
        return bspline_field

    elif field_type == 'exponential':
        libfn = get_lib_fn("simulateExponentialDisplacementField%iD" % image_dimension)
        field = libfn(domain_image.pointer, number_of_random_points, sd_noise, 
                      enforce_stationary_boundary, sd_smoothing)
        exp_field = ants.from_pointer(field).clone('float')
        return exp_field

    else:  
        raise ValueError("Unrecognized field type.")




 

__all__ = ['label_clusters']

import ants
from ants.internal import get_lib_fn, process_arguments
from ants.decorators import image_method

@image_method
def label_clusters(image, min_cluster_size=50, min_thresh=1e-6, max_thresh=1, fully_connected=False):
    """
    This will give a unique ID to each connected 
    component 1 through N of size > min_cluster_size

    ANTsR function: `labelClusters`

    Arguments
    ---------
    image : ANTsImage 
        input image e.g. a statistical map
    
    min_cluster_size : integer  
        throw away clusters smaller than this value
    
    min_thresh : scalar
        threshold to a statistical map
    
    max_thresh : scalar
        threshold to a statistical map
    
    fully_connected : boolean
        boolean sets neighborhood connectivity pattern
    
    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data('r16') )
    >>> timageFully = ants.label_clusters( image, 10, 128, 150, True )
    >>> timageFace = ants.label_clusters( image, 10, 128, 150, False )
    """
    dim = image.dimension
    clust = ants.threshold_image(image, min_thresh, max_thresh)
    temp = int(fully_connected)
    args = [dim, clust, clust, min_cluster_size, temp]
    processed_args = process_arguments(args)
    libfn = get_lib_fn('LabelClustersUniquely')
    libfn(processed_args)
    return clust



__all__ = ["get_centroids"]

import numpy as np
import ants
from ants.decorators import image_method

@image_method
def get_centroids(image, clustparam=0):
    """
    Reduces a variate/statistical/network image to a set of centroids
    describing the center of each stand-alone non-zero component in the image

    ANTsR function: `getCentroids`

    Arguments
    ---------
    image : ANTsImage
        image from which centroids will be calculated

    clustparam : integer
        look at regions greater than or equal to this size

    Returns
    -------
    ndarray

    Example
    -------
    >>> import ants
    >>> image = ants.image_read( ants.get_ants_data( "r16" ) )
    >>> image = ants.threshold_image( image, 90, 120 )
    >>> image = ants.label_clusters( image, 10 )
    >>> cents = ants.get_centroids( image )
    """
    imagedim = image.dimension
    if clustparam > 0:
        mypoints = ants.label_clusters(image, clustparam, max_thresh=1e15)
    if clustparam == 0:
        mypoints = image.clone()
    mypoints = ants.label_stats(mypoints, mypoints)
    nonzero = mypoints[["LabelValue"]] > 0
    mypoints = mypoints[nonzero["LabelValue"]]
    mypoints = mypoints.iloc[:, :]
    x = mypoints.x
    y = mypoints.y

    if imagedim == 3:
        z = mypoints.z
    else:
        z = np.zeros(mypoints.shape[0])

    if imagedim == 4:
        t = mypoints.t
    else:
        t = np.zeros(mypoints.shape[0])

    centroids = np.stack([x, y, z, t]).T
    return centroids





__all__ = ['ANTsImage',
           'copy_image_info',
           'set_origin',
           'get_origin',
           'set_direction',
           'get_direction',
           'set_spacing',
           'get_spacing',
           'is_image',
           'from_pointer']

import os

import numpy as np
import pandas as pd

from functools import partialmethod
import inspect

import ants
from ants.internal import get_lib_fn

_supported_ptypes = {'unsigned char', 'unsigned int', 'float', 'double'}
_supported_dtypes = {'uint8', 'uint32', 'float32', 'float64'}
_itk_to_npy_map = {
    'unsigned char': 'uint8',
    'unsigned int': 'uint32',
    'float': 'float32',
    'double': 'float64'}
_npy_to_itk_map = {
    'uint8': 'unsigned char',
    'uint32':'unsigned int',
    'float32': 'float',
    'float64': 'double'}


class ANTsImage(object):

    def __init__(self, pointer):
        """
        Initialize an ANTsImage.

        Creating an ANTsImage requires a pointer to an underlying ITK image that
        is stored via a nanobind class wrapping a AntsImage struct.

        Arguments
        ---------
        pointer : nb::class
            nanobind class wrapping the struct holding the pointer to the underlying ITK image object

        """
        self.pointer = pointer
        self.channels_first = False
        self._array = None

    @property
    def _libsuffix(self):
        return str(type(self.pointer)).split('AntsImage')[-1].split("'")[0]

    @property
    def shape(self):
        return tuple(get_lib_fn('getShape')(self.pointer))

    @property
    def physical_shape(self):
        return tuple([round(sh*sp,3) for sh,sp in zip(self.shape, self.spacing)])

    @property
    def is_rgb(self):
        return 'RGB' in self._libsuffix

    @property
    def has_components(self):
        suffix = self._libsuffix
        return suffix.startswith('V') or suffix.startswith('RGB')

    @property
    def components(self):
        if not self.has_components:
            return 1

        return get_lib_fn('getComponents')(self.pointer)

    @property
    def pixeltype(self):
        ptype = self._libsuffix[:-1]
        if self.has_components:
            if self.is_rgb:
                ptype = ptype[3:]
            else:
                ptype = ptype[1:]

        ptype_map = {'UC': 'unsigned char',
                     'UI': 'unsigned int',
                     'F': 'float',
                     'D': 'double'}
        return ptype_map[ptype]

    @property
    def dtype(self):
        return _itk_to_npy_map[self.pixeltype]

    @property
    def dimension(self):
        return int(self._libsuffix[-1])

    @property
    def spacing(self):
        """
        Get image spacing

        Returns
        -------
        tuple
        """
        return tuple(get_lib_fn('getSpacing')(self.pointer))

    def set_spacing(self, new_spacing):
        """
        Set image spacing

        Arguments
        ---------
        new_spacing : tuple or list
            updated spacing for the image.
            should have one value for each dimension

        Returns
        -------
        None
        """
        if not isinstance(new_spacing, (tuple, list)):
            raise ValueError('arg must be tuple or list')
        if len(new_spacing) != self.dimension:
            raise ValueError('must give a spacing value for each dimension (%i)' % self.dimension)

        get_lib_fn('setSpacing')(self.pointer, new_spacing)

    @property
    def origin(self):
        """
        Get image origin

        Returns
        -------
        tuple
        """
        return tuple(get_lib_fn('getOrigin')(self.pointer))

    def set_origin(self, new_origin):
        """
        Set image origin

        Arguments
        ---------
        new_origin : tuple or list
            updated origin for the image.
            should have one value for each dimension

        Returns
        -------
        None
        """
        if not isinstance(new_origin, (tuple, list)):
            raise ValueError('arg must be tuple or list')
        if len(new_origin) != self.dimension:
            raise ValueError('must give a origin value for each dimension (%i)' % self.dimension)

        get_lib_fn('setOrigin')(self.pointer, new_origin)

    @property
    def direction(self):
        """
        Get image direction

        Returns
        -------
        tuple
        """
        return np.array(get_lib_fn('getDirection')(self.pointer)).reshape(self.dimension,self.dimension)

    def set_direction(self, new_direction):
        """
        Set image direction

        Arguments
        ---------
        new_direction : numpy.ndarray or tuple or list
            updated direction for the image.
            should have one value for each dimension

        Returns
        -------
        None
        """
        if isinstance(new_direction, (tuple,list)):
            new_direction = np.asarray(new_direction)

        if not isinstance(new_direction, np.ndarray):
            raise ValueError('arg must be np.ndarray or tuple or list')
        if len(new_direction) != self.dimension:
            raise ValueError('must give a origin value for each dimension (%i)' % self.dimension)

        get_lib_fn('setDirection')(self.pointer, new_direction)

    @property
    def orientation(self):
        if self.dimension == 3:
            return self.get_orientation()
        else:
            return None

    def view(self, single_components=False):
        """
        Geet a numpy array providing direct, shared access to the image data.
        IMPORTANT: If you alter the view, then the underlying image data
        will also be altered.

        Arguments
        ---------
        single_components : boolean (default is False)
            if True, keep the extra component dimension in returned array even
            if image only has one component (i.e. self.has_components == False)

        Returns
        -------
        ndarray
        """
        if self.is_rgb:
            img = self.rgb_to_vector()
        else:
            img = self
        dtype = img.dtype
        shape = img.shape[::-1]
        if img.has_components or (single_components == True):
            shape = list(shape) + [img.components]
        memview = get_lib_fn('toNumpy')(img.pointer)
        return np.asarray(memview).view(dtype = dtype).reshape(shape).view(np.ndarray).T

    def numpy(self, single_components=False):
        """
        Get a numpy array copy representing the underlying image data. Altering
        this ndarray will have NO effect on the underlying image data.

        Arguments
        ---------
        single_components : boolean (default is False)
            if True, keep the extra component dimension in returned array even
            if image only has one component (i.e. self.has_components == False)

        Returns
        -------
        ndarray
        """
        array = np.array(self.view(single_components=single_components), copy=True, dtype=self.dtype)
        if self.has_components or (single_components == True):
            if not self.channels_first:
                array = np.rollaxis(array, 0, self.dimension+1)
        return array

    def astype(self, dtype):
        """
        Cast & clone an ANTsImage to a given numpy datatype.

        Map:
            uint8   : unsigned char
            uint32  : unsigned int
            float32 : float
            float64 : double
        """
        if dtype not in _supported_dtypes:
            raise ValueError('Datatype %s not supported. Supported types are %s' % (dtype, _supported_dtypes))

        pixeltype = _npy_to_itk_map[dtype]
        return self.clone(pixeltype)

    def to_file(self, filename):
        """
        Write the ANTsImage to file

        Args
        ----
        filename : string
            filepath to which the image will be written
        """
        filename = os.path.expanduser(filename)
        get_lib_fn('toFile')(self.pointer, filename)
    to_filename = to_file

    def apply(self, fn):
        """
        Apply an arbitrary function to ANTsImage.

        Args
        ----
        fn : python function or lambda
            function to apply to ENTIRE image at once

        Returns
        -------
        ANTsImage
            image with function applied to it
        """
        this_array = self.numpy()
        new_array = fn(this_array)
        return self.new_image_like(new_array)

    ## NUMPY FUNCTIONS ##
    def abs(self, axis=None):
        """ Return absolute value of image """
        return np.abs(self.numpy())
    def mean(self, axis=None):
        """ Return mean along specified axis """
        return self.numpy().mean(axis=axis)
    def median(self, axis=None):
        """ Return median along specified axis """
        return np.median(self.numpy(), axis=axis)
    def std(self, axis=None):
        """ Return std along specified axis """
        return self.numpy().std(axis=axis)
    def sum(self, axis=None, keepdims=False):
        """ Return sum along specified axis """
        return self.numpy().sum(axis=axis, keepdims=keepdims)
    def min(self, axis=None):
        """ Return min along specified axis """
        return self.numpy().min(axis=axis)
    def max(self, axis=None):
        """ Return max along specified axis """
        return self.numpy().max(axis=axis)
    def range(self, axis=None):
        """ Return range tuple along specified axis """
        return (self.min(axis=axis), self.max(axis=axis))
    def argmin(self, axis=None):
        """ Return argmin along specified axis """
        return self.numpy().argmin(axis=axis)
    def argmax(self, axis=None):
        """ Return argmax along specified axis """
        return self.numpy().argmax(axis=axis)
    def argrange(self, axis=None):
        """ Return argrange along specified axis """
        amin = self.argmin(axis=axis)
        amax = self.argmax(axis=axis)
        if axis is None:
            return (amin, amax)
        else:
            return np.stack([amin, amax]).T
    def flatten(self):
        """ Flatten image data """
        return self.numpy().flatten()
    def nonzero(self):
        """ Return non-zero indices of image """
        return self.numpy().nonzero()
    def unique(self, sort=False):
        """ Return unique set of values in image """
        unique_vals = np.unique(self.numpy())
        if sort:
            unique_vals = np.sort(unique_vals)
        return unique_vals

    ## OVERLOADED OPERATORS ##
    def __add__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array + other
        return self.new_image_like(new_array)

    __radd__ = __add__

    def __sub__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array - other
        return self.new_image_like(new_array)

    def __rsub__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = other - this_array
        return self.new_image_like(new_array)

    def __mul__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array * other
        return self.new_image_like(new_array)

    __rmul__ = __mul__

    def __truediv__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array / other
        return self.new_image_like(new_array)

    def __pow__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array ** other
        return self.new_image_like(new_array)

    def __gt__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array > other
        return self.new_image_like(new_array.astype('uint8'))

    def __ge__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array >= other
        return self.new_image_like(new_array.astype('uint8'))

    def __lt__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array < other
        return self.new_image_like(new_array.astype('uint8'))

    def __le__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array <= other
        return self.new_image_like(new_array.astype('uint8'))

    def __eq__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array == other
        return self.new_image_like(new_array.astype('uint8'))

    def __ne__(self, other):
        this_array = self.numpy()

        if is_image(other):
            if not ants.image_physical_space_consistency(self, other):
                raise ValueError('images do not occupy same physical space')
            other = other.numpy()

        new_array = this_array != other
        return self.new_image_like(new_array.astype('uint8'))

    def __getitem__(self, idx):
        if self.has_components:
            return ants.merge_channels([
                img[idx] for img in ants.split_channels(self)
            ])

        if isinstance(idx, ANTsImage):
            if not ants.image_physical_space_consistency(self, idx):
                raise ValueError('images do not occupy same physical space')
            return self.numpy().__getitem__(idx.numpy().astype('bool'))

        # convert idx to tuple if it is not, eg im[10] or im[10:20]
        if not isinstance(idx, tuple):
            idx = (idx,)

        ndim = len(self.shape)

        if len(idx) > ndim:
            raise ValueError('Too many indices for image')
        if len(idx) < ndim:
            # If not all dimensions are indexed, assume the rest are full slices
            # eg im[10] -> im[10, :, :]
            idx = idx + (slice(None),) * (ndim - len(idx))

        sizes = list(self.shape)
        starts = [0] * ndim
        stops = list(self.shape)
        for i in range(ndim):
            ti = idx[i]
            if isinstance(ti, slice):
                if ti.start:
                    starts[i] = ti.start
                if ti.stop:
                    if ti.stop < 0:
                        stops[i] = self.shape[i] + ti.stop
                    else:
                        stops[i] = ti.stop

                sizes[i] = stops[i] - starts[i]

                if stops[i] < starts[i]:
                    raise ValueError('Reverse indexing is not supported.')

            elif isinstance(ti, int):
                starts[i] = ti
                sizes[i] = 0

            if sizes[i] == 0:
                ndim -= 1

        if ndim < 2:
            return self.numpy().__getitem__(idx)

        libfn = get_lib_fn('getItem%i' % ndim)
        new_ptr = libfn(self.pointer, starts, sizes)
        new_image = from_pointer(new_ptr)
        return new_image

    def __setitem__(self, idx, value):
        arr = self.view()

        if is_image(value):
            value = value.numpy()

        if is_image(idx):
            if not ants.image_physical_space_consistency(self, idx):
                raise ValueError('images do not occupy same physical space')
            arr.__setitem__(idx.numpy().astype('bool'), value)
        else:
            arr.__setitem__(idx, value)


    def __iter__(self):
        # Do not allow iteration on ANTsImage. Builtin iteration, eg sum(), will generally be much slower
        # than using numpy methods. We need to explicitly disallow it to prevent breaking object state.
        raise TypeError("ANTsImage is not iterable. See docs for available functions, or use numpy.")


    def __repr__(self):
        if self.dimension == 3:
            s = 'ANTsImage ({})\n'.format(self.orientation)
        else:
            s = 'ANTsImage\n'
        s = s +\
            '\t {:<10} : {} ({})\n'.format('Pixel Type', self.pixeltype, self.dtype)+\
            '\t {:<10} : {}{}\n'.format('Components', self.components, ' (RGB)' if 'RGB' in self._libsuffix else '')+\
            '\t {:<10} : {}\n'.format('Dimensions', self.shape)+\
            '\t {:<10} : {}\n'.format('Spacing', tuple([round(s,4) for s in self.spacing]))+\
            '\t {:<10} : {}\n'.format('Origin', tuple([round(o,4) for o in self.origin]))+\
            '\t {:<10} : {}\n'.format('Direction', np.round(self.direction.flatten(),4))
        return s

    def __getstate__(self):
        """
        import ants
        import pickle
        import numpy as np
        from copy import deepcopy
        img = ants.image_read( ants.get_ants_data("r16"))
        img_pickled = pickle.dumps(img)
        img2 = pickle.loads(img_pickled)
        img3 = deepcopy(img)
        img += 10
        print(img.mean(), img3.mean())
        """
        return self.numpy(), self.origin, self.spacing, self.direction, self.has_components, self.is_rgb

    def __setstate__(self, state):
        data, origin, spacing, direction, has_components, is_rgb = state
        image = ants.from_numpy(np.copy(data), origin=origin, spacing=spacing, direction=direction, has_components=has_components, is_rgb=is_rgb)
        self.__dict__ = image.__dict__

def copy_image_info(reference, target):
    """
    Copy origin, direction, and spacing from one antsImage to another

    ANTsR function: `antsCopyImageInfo`

    Arguments
    ---------
    reference : ANTsImage
        Image to get values from.
    target  : ANTsImAGE
        Image to copy values to

    Returns
    -------
    ANTsImage
        Target image with reference header information
    """
    target.set_origin(reference.origin)
    target.set_direction(reference.direction)
    target.set_spacing(reference.spacing)
    return target


def set_origin(image, origin):
    """
    Set origin of ANTsImage

    ANTsR function: `antsSetOrigin`
    """
    image.set_origin(origin)

def get_origin(image):
    """
    Get origin of ANTsImage

    ANTsR function: `antsGetOrigin`
    """

    return image.origin

def set_direction(image, direction):
    """
    Set direction of ANTsImage

    ANTsR function: `antsSetDirection`
    """
    image.set_direction(direction)

def get_direction(image):
    """
    Get direction of ANTsImage

    ANTsR function: `antsGetDirection`
    """
    return image.direction

def set_spacing(image, spacing):
    """
    Set spacing of ANTsImage

    ANTsR function: `antsSetSpacing`
    """
    image.set_spacing(spacing)

def get_spacing(image):
    """
    Get spacing of ANTsImage

    ANTsR function: `antsGetSpacing`
    """
    return image.spacing


def is_image(object):
    return isinstance(object, ANTsImage)

def from_pointer(pointer):
    return ANTsImage(pointer)



__all__ = ['slice_image']

import math
import numpy as np
import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def slice_image(image, axis, idx, collapse_strategy=0):
    """
    Slice an image.

    Arguments
    ---------
    axis: integer 
        Which axis.

    idx: integer
        Which slice number.    

    collapse_strategy:  integer
        Collapse strategy for sub-matrix: 0, 1, or 2.  0: collapse to sub-matrix 
        if positive-definite.  Otherwise throw an exception. Default.  1: Collapse 
        to identity.  2:  Collapse to sub-matrix if positive definite. Otherwise
        collapse to identity.

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> mni2 = ants.slice_image(mni, axis=1, idx=100)
    """
    if image.has_components:
        ilist = ants.split_channels(image)
        if image.dimension == 2:
            return np.stack(tuple([i.slice_image(axis, idx, collapse_strategy) for i in ilist]), axis=-1)
        else:
            return ants.merge_channels([i.slice_image(axis, idx, collapse_strategy) for i in ilist])
    
    if axis == -1:
        axis = image.dimension - 1
        
    if axis > (image.dimension - 1) or axis < 0:
        raise Exception('The axis must be between 0 and image.dimension - 1')
        
    if image.dimension == 2:
        if axis == 0:
            return image[idx,:]
        elif axis == 1:
            return image[:,idx]
        raise Exception('Parameters not understood for 2D image.')
        
    if collapse_strategy != 0 and collapse_strategy != 1 and collapse_strategy != 2:
        raise ValueError('collapse_strategy must be 0, 1, or 2.') 

    inpixeltype = image.pixeltype
    ndim = image.dimension
    if image.pixeltype != 'float':
        image = image.clone('float')

    libfn = get_lib_fn('sliceImage')
    itkimage = libfn(image.pointer, axis, idx, collapse_strategy)

    return ants.from_pointer(itkimage).clone(inpixeltype)





"""
Affine transforms

See http://www.cs.cornell.edu/courses/cs4620/2010fa/lectures/03transforms3D.pdf
"""

__all__ = [
    "Zoom2D",
    "RandomZoom2D",
    "Rotate2D",
    "RandomRotate2D",
    "Shear2D",
    "RandomShear2D",
    "Translate2D",
    "RandomTranslate2D",
]

import random
import math
import numpy as np

from ...core import ants_transform as tio


class Translate2D(object):
    """
    Create an ANTs Affine Transform with a specified translation.
    """

    def __init__(self, translation, reference=None, lazy=False):
        """
        Initialize a Translate2D object

        Arguments
        ---------
        translation : list or tuple
            translation values for each axis, in degrees.
            Negative values can be used for translation in the
            other direction

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(translation, (list, tuple))) or (len(translation) != 2):
            raise ValueError("translation argument must be list/tuple with two values!")

        self.translation = translation
        self.lazy = lazy
        self.reference = reference

        self.tx = tio.ANTsTransform(
            precision="float", dimension=2, transform_type="AffineTransform"
        )
        if self.reference is not None:
            self.tx.set_fixed_parameters(self.reference.get_center_of_mass())

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with the given
        translation parameters.  Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.Translate2D(translation=(10,0))
        >>> img2_x = tx.transform(img)
        >>> tx = ants.contrib.Translate2D(translation=(-10,0)) # other direction
        >>> img2_x = tx.transform(img)
        >>> tx = ants.contrib.Translate2D(translation=(0,10))
        >>> img2_z = tx.transform(img)
        >>> tx = ants.contrib.Translate2D(translation=(10,10))
        >>> img2 = tx.transform(img)
        """
        # convert to radians and unpack
        translation_x, translation_y = self.translation

        translation_matrix = np.array([[1, 0, translation_x], [0, 1, translation_y]])
        self.tx.set_parameters(translation_matrix)
        if self.lazy or X is None:
            return self.tx
        else:
            if y is None:
                return self.tx.apply_to_image(X, reference=self.reference)
            else:
                return (
                    self.tx.apply_to_image(X, reference=self.reference),
                    self.tx.apply_to_image(y, reference=self.reference),
                )


class RandomTranslate2D(object):
    """
    Apply a Translate2D transform to an image, but with the
    parameters randomly generated from a user-specified range.
    The range is determined by a mean (first parameter) and standard deviation
    (second parameter) via calls to random.gauss.
    """

    def __init__(self, translation_range, reference=None, lazy=False):
        """
        Initialize a RandomTranslate2D object

        Arguments
        ---------
        translation_range : list or tuple
            Lower and Upper bounds on rotation parameter, in degrees.
            e.g. translation_range = (-10,10) will result in a random
            draw of the rotation parameters between -10 and 10 degrees

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(translation_range, (list, tuple))) or (
            len(translation_range) != 2
        ):
            raise ValueError("shear_range argument must be list/tuple with two values!")

        self.translation_range = translation_range
        self.reference = reference
        self.lazy = lazy

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with
        translation parameters randomly generated from the user-specified
        range.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.RandomShear2D(translation_range=(-10,10))
        >>> img2 = tx.transform(img)
        """
        # random draw in translation range
        translation_x = random.gauss(
            self.translation_range[0], self.translation_range[1]
        )
        translation_y = random.gauss(
            self.translation_range[0], self.translation_range[1]
        )
        self.params = (translation_x, translation_y)

        tx = Translate2D(
            (translation_x, translation_y), reference=self.reference, lazy=self.lazy
        )

        return tx.transform(X, y)


class Shear2D(object):
    """
    Create an ANTs Affine Transform with a specified shear.
    """

    def __init__(self, shear, reference=None, lazy=False):
        """
        Initialize a Shear2D object

        Arguments
        ---------
        shear : list or tuple
            shear values for each axis, in degrees.
            Negative values can be used for shear in the
            other direction

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(shear, (list, tuple))) or (len(shear) != 2):
            raise ValueError("shear argument must be list/tuple with two values!")

        self.shear = shear
        self.lazy = lazy
        self.reference = reference

        self.tx = tio.ANTsTransform(
            precision="float", dimension=2, transform_type="AffineTransform"
        )
        if self.reference is not None:
            self.tx.set_fixed_parameters(self.reference.get_center_of_mass())

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with the given
        shear parameters.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.Shear2D(shear=(10,0,0))
        >>> img2_x = tx.transform(img)# x axis stays same
        >>> tx = ants.contrib.Shear2D(shear=(-10,0,0)) # other direction
        >>> img2_x = tx.transform(img)# x axis stays same
        >>> tx = ants.contrib.Shear2D(shear=(0,10,0))
        >>> img2_y = tx.transform(img) # y axis stays same
        >>> tx = ants.contrib.Shear2D(shear=(0,0,10))
        >>> img2_z = tx.transform(img) # z axis stays same
        >>> tx = ants.contrib.Shear2D(shear=(10,10,10))
        >>> img2 = tx.transform(img)
        """
        # convert to radians and unpack
        shear = [math.pi / 180 * s for s in self.shear]
        shear_x, shear_y = shear

        shear_matrix = np.array([[1, shear_x, 0], [shear_y, 1, 0]])
        self.tx.set_parameters(shear_matrix)
        if self.lazy or X is None:
            return self.tx
        else:
            if y is None:
                return self.tx.apply_to_image(X, reference=self.reference)
            else:
                return (
                    self.tx.apply_to_image(X, reference=self.reference),
                    self.tx.apply_to_image(y, reference=self.reference),
                )


class RandomShear2D(object):
    """
    Apply a Shear2D transform to an image, but with the shear
    parameters randomly generated from a user-specified range.
    The range is determined by a mean (first parameter) and standard deviation
    (second parameter) via calls to random.gauss.
    """

    def __init__(self, shear_range, reference=None, lazy=False):
        """
        Initialize a RandomShear2D object

        Arguments
        ---------
        shear_range : list or tuple
            Lower and Upper bounds on rotation parameter, in degrees.
            e.g. shear_range = (-10,10) will result in a random
            draw of the rotation parameters between -10 and 10 degrees

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(shear_range, (list, tuple))) or (len(shear_range) != 2):
            raise ValueError("shear_range argument must be list/tuple with two values!")

        self.shear_range = shear_range
        self.reference = reference
        self.lazy = lazy

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with
        shear parameters randomly generated from the user-specified
        range.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.RandomShear2D(shear_range=(-10,10))
        >>> img2 = tx.transform(img)
        """
        # random draw in shear range
        shear_x = random.gauss(self.shear_range[0], self.shear_range[1])
        shear_y = random.gauss(self.shear_range[0], self.shear_range[1])
        self.params = (shear_x, shear_y)

        tx = Shear2D((shear_x, shear_y), reference=self.reference, lazy=self.lazy)

        return tx.transform(X, y)


class Rotate2D(object):
    """
    Create an ANTs Affine Transform with a specified level
    of rotation.
    """

    def __init__(self, rotation, reference=None, lazy=False):
        """
        Initialize a Rotate2D object

        Arguments
        ---------
        rotation : scalar
            rotation value in degrees.
            Negative values can be used for rotation in the
            other direction

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        self.rotation = rotation
        self.lazy = lazy
        self.reference = reference

        self.tx = tio.ANTsTransform(
            precision="float", dimension=2, transform_type="AffineTransform"
        )
        if self.reference is not None:
            self.tx.set_fixed_parameters(self.reference.get_center_of_mass())

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with the given
        rotation parameters.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.Rotate2D(rotation=(10,-5,12))
        >>> img2 = tx.transform(img)
        """
        # unpack zoom range
        rotation = self.rotation

        # Rotation about X axis
        theta = math.pi / 180 * rotation
        rotation_matrix = np.array(
            [[np.cos(theta), -np.sin(theta), 0], [np.sin(theta), np.cos(theta), 0]]
        )

        self.tx.set_parameters(rotation_matrix)
        if self.lazy or X is None:
            return self.tx
        else:
            if y is None:
                return self.tx.apply_to_image(X, reference=self.reference)
            else:
                return (
                    self.tx.apply_to_image(X, reference=self.reference),
                    self.tx.apply_to_image(y, reference=self.reference),
                )


class RandomRotate2D(object):
    """
    Apply a Rotated2D transform to an image, but with the zoom
    parameters randomly generated from a user-specified range.
    The range is determined by a mean (first parameter) and standard deviation
    (second parameter) via calls to random.gauss.
    """

    def __init__(self, rotation_range, reference=None, lazy=False):
        """
        Initialize a RandomRotate2D object

        Arguments
        ---------
        rotation_range : list or tuple
            Lower and Upper bounds on rotation parameter, in degrees.
            e.g. rotation_range = (-10,10) will result in a random
            draw of the rotation parameters between -10 and 10 degrees

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(rotation_range, (list, tuple))) or (
            len(rotation_range) != 2
        ):
            raise ValueError(
                "rotation_range argument must be list/tuple with two values!"
            )

        self.rotation_range = rotation_range
        self.reference = reference
        self.lazy = lazy

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with
        rotation parameters randomly generated from the user-specified
        range.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.RandomRotate2D(rotation_range=(-10,10))
        >>> img2 = tx.transform(img)
        """
        # random draw in rotation range
        rotation = random.gauss(self.rotation_range[0], self.rotation_range[1])
        self.params = rotation

        tx = Rotate2D(rotation, reference=self.reference, lazy=self.lazy)

        return tx.transform(X, y)


class Zoom2D(object):
    """
    Create an ANTs Affine Transform with a specified level
    of zoom. Any value greater than 1 implies a "zoom-out" and anything
    less than 1 implies a "zoom-in".
    """

    def __init__(self, zoom, reference=None, lazy=False):
        """
        Initialize a Zoom2D object

        Arguments
        ---------
        zoom_range : list or tuple
            Lower and Upper bounds on zoom parameter.
            e.g. zoom_range = (0.7,0.9) will result in a random
            draw of the zoom parameters between 0.7 and 0.9

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(zoom, (list, tuple))) or (len(zoom) != 2):
            raise ValueError("zoom_range argument must be list/tuple with two values!")

        self.zoom = zoom
        self.lazy = lazy
        self.reference = reference

        self.tx = tio.ANTsTransform(
            precision="float", dimension=2, transform_type="AffineTransform"
        )
        if self.reference is not None:
            self.tx.set_fixed_parameters(self.reference.get_center_of_mass())

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with the given
        zoom parameters.   Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.Zoom2D(zoom=(0.8,0.8,0.8))
        >>> img2 = tx.transform(img)
        """
        # unpack zoom range
        zoom_x, zoom_y = self.zoom

        self.params = (zoom_x, zoom_y)
        zoom_matrix = np.array([[zoom_x, 0, 0], [0, zoom_y, 0]])
        self.tx.set_parameters(zoom_matrix)
        if self.lazy or X is None:
            return self.tx
        else:
            if y is None:
                return self.tx.apply_to_image(X, reference=self.reference)
            else:
                return (
                    self.tx.apply_to_image(X, reference=self.reference),
                    self.tx.apply_to_image(y, reference=self.reference),
                )


class RandomZoom2D(object):
    """
    Apply a Zoom2D transform to an image, but with the zoom
    parameters randomly generated from a user-specified range.
    The range is determined by a mean (first parameter) and standard deviation
    (second parameter) via calls to random.gauss.
    """

    def __init__(self, zoom_range, reference=None, lazy=False):
        """
        Initialize a RandomZoom2D object

        Arguments
        ---------
        zoom_range : list or tuple
            Lower and Upper bounds on zoom parameter.
            e.g. zoom_range = (0.7,0.9) will result in a random
            draw of the zoom parameters between 0.7 and 0.9

        reference : ANTsImage (optional but recommended)
            image providing the reference space for the transform.
            this will also set the transform fixed parameters.

        lazy : boolean (default = False)
            if True, calling the `transform` method only returns
            the randomly generated transform and does not actually
            transform the image
        """
        if (not isinstance(zoom_range, (list, tuple))) or (len(zoom_range) != 2):
            raise ValueError("zoom_range argument must be list/tuple with two values!")

        self.zoom_range = zoom_range
        self.reference = reference
        self.lazy = lazy

    def transform(self, X=None, y=None):
        """
        Transform an image using an Affine transform with
        zoom parameters randomly generated from the user-specified
        range.  Return the transform if X=None.

        Arguments
        ---------
        X : ANTsImage
            Image to transform

        y : ANTsImage (optional)
            Another image to transform

        Returns
        -------
        ANTsImage if y is None, else a tuple of ANTsImage types

        Examples
        --------
        >>> import ants
        >>> img = ants.image_read(ants.get_data('r16'))
        >>> tx = ants.contrib.RandomZoom2D(zoom_range=(0.8,0.9))
        >>> img2 = tx.transform(img)
        """
        # random draw in zoom range
        zoom_x = np.exp(
            random.gauss(np.log(self.zoom_range[0]), np.log(self.zoom_range[1]))
        )
        zoom_y = np.exp(
            random.gauss(np.log(self.zoom_range[0]), np.log(self.zoom_range[1]))
        )
        self.params = (zoom_x, zoom_y)

        tx = Zoom2D((zoom_x, zoom_y), reference=self.reference, lazy=self.lazy)

        return tx.transform(X, y)






__all__ = ['morphology']

import ants
from ants.decorators import image_method

@image_method
def morphology(image, operation, radius, mtype='binary', value=1,
               shape='ball', radius_is_parametric=False, thickness=1,
               lines=3, include_center=False):
    """
    Apply morphological operations to an image

    ANTsR function: `morphology`

    Arguments
    ---------
    input : ANTsImage
        input image

    operation : string
        operation to apply
            "close" Morpholgical closing
            "dilate" Morpholgical dilation
            "erode" Morpholgical erosion
            "open" Morpholgical opening

    radius : scalar
        radius of structuring element

    mtype : string
        type of morphology
            "binary" Binary operation on a single value
            "grayscale" Grayscale operations

    value : scalar
        value to operation on (type='binary' only)

    shape : string
        shape of the structuring element ( type='binary' only )
            "ball" spherical structuring element
            "box" box shaped structuring element
            "cross" cross shaped structuring element
            "annulus" annulus shaped structuring element
            "polygon" polygon structuring element

    radius_is_parametric : boolean
        used parametric radius boolean (shape='ball' and shape='annulus' only)

    thickness : scalar
        thickness (shape='annulus' only)

    lines : integer
        number of lines in polygon (shape='polygon' only)

    include_center : boolean
        include center of annulus boolean (shape='annulus' only)

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16') , 2 )
    >>> mask = ants.get_mask( fi )
    >>> dilated_ball = ants.morphology( mask, operation='dilate', radius=3, mtype='binary', shape='ball')
    >>> eroded_box = ants.morphology( mask, operation='erode', radius=3, mtype='binary', shape='box')
    >>> opened_annulus = ants.morphology( mask, operation='open', radius=5, mtype='binary', shape='annulus', thickness=2)
    """
    if image.components > 1:
        raise ValueError('multichannel images not yet supported')

    _sflag_dict = {'ball': 1, 'box': 2, 'cross': 3, 'annulus': 4, 'polygon': 5}
    sFlag = _sflag_dict.get(shape, 0)

    if sFlag == 0:
        raise ValueError('invalid element shape')

    radius_is_parametric = radius_is_parametric * 1
    include_center = include_center * 1
    if (mtype == 'binary'):
        if (operation == 'dilate'):
            if (sFlag == 5):
                ret = ants.iMath(image, 'MD', radius, value, sFlag, lines)
            else:
                ret = ants.iMath(image, 'MD', radius, value, sFlag, radius_is_parametric, thickness, include_center)
        elif (operation == 'erode'):
            if (sFlag == 5):
                ret = ants.iMath(image, 'ME', radius, value, sFlag, lines)
            else:
                ret = ants.iMath(image, 'ME', radius, value, sFlag, radius_is_parametric, thickness, include_center)
        elif (operation == 'open'):
            if (sFlag == 5):
                ret = ants.iMath(image, 'MO', radius, value, sFlag, lines)
            else:
                ret = ants.iMath(image, 'MO', radius, value, sFlag, radius_is_parametric, thickness, include_center)
        elif (operation == 'close'):
            if (sFlag == 5):
                ret = ants.iMath(image, 'MC', radius, value, sFlag, lines)
            else:
                ret = ants.iMath(image, 'MC', radius, value, sFlag, radius_is_parametric, thickness, include_center)
        else:
            raise ValueError('Invalid morphology operation')
    elif (mtype == 'grayscale'):
        if (operation == 'dilate'):
            ret = ants.iMath(image, 'GD', radius)
        elif (operation == 'erode'):
            ret = ants.iMath(image, 'GE', radius)
        elif (operation == 'open'):
            ret = ants.iMath(image, 'GO', radius)
        elif (operation == 'close'):
            ret = ants.iMath(image, 'GC', radius)
        else:
            raise ValueError('Invalid morphology operation')
    else:
        raise ValueError('Invalid morphology type')

    return ret




 
__all__ = ['otsu_segmentation']

def otsu_segmentation(image, k, mask=None):
    """
    Otsu image segmentation

    This is a very fast segmentation algorithm good for quick explortation, 
    but does not return probability maps.

    ANTsR function: `thresholdImage(image, 'Otsu', k)`
    
    Arguments
    ---------
    image : ANTsImage 
        input image
    
    k : integer
        integer number of classes. Note that a background class will 
        be added to this, so the resulting segmentation will 
        have k+1 unique values.
    
    mask : ANTsImage 
        segment inside this mask

    Returns
    -------
    ANTsImage

    Example
    -------
    >>> import ants
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> seg = mni.otsu_segmentation(k=3) #0=bg,1=csf,2=gm,3=wm
    """
    if mask is not None:
        image = image.mask_image(mask)

    seg = image.threshold_image('Otsu', k)
    return seg







__all__ = ['label_geometry_measures']

from tempfile import mktemp
import pandas as pd
import numpy as np

from ants.internal import get_lib_fn, process_arguments
from ants.decorators import image_method

@image_method
def label_geometry_measures(label_image, intensity_image=None):
    """
    Wrapper for the ANTs funtion LabelGeometryMeasures

    ANTsR function: `labelGeometryMeasures`

    Arguments
    ---------
    label_image : ANTsImage
        image on which to compute geometry. Labels must be representable as uint32.
    intensity_image : ANTsImage (optional)
        image with intensity values

    Returns
    -------
    pandas.DataFrame

    Example
    -------
    >>> import ants
    >>> fi = ants.image_read( ants.get_ants_data('r16') )
    >>> seg = ants.kmeans_segmentation( fi, 3 )['segmentation']
    >>> geom = ants.label_geometry_measures(seg,fi)
    """
    if intensity_image is None:
        intensity_image = label_image.clone()

    outcsv = mktemp(suffix='.csv')
    # Library function requires unsigned int labels
    if label_image.pixeltype != 'unsigned int':
        label_image_int = label_image.clone('unsigned int')
        if not np.all(label_image.numpy() == label_image_int.numpy()):
            raise ValueError('Input label values must be representable as uint32.')
        label_image = label_image.clone('unsigned int')
    veccer = [label_image.dimension, label_image_int, intensity_image, outcsv]
    veccer_processed = process_arguments(veccer)
    libfn = get_lib_fn('LabelGeometryMeasures')
    pp = libfn(veccer_processed)
    pp = pd.read_csv(outcsv)
    if 'VolumeInVoxels' in pp.columns and not 'VolumeInMillimeters' in pp.columns:
        spc = np.prod(label_image.spacing)
        pp['VolumeInMillimeters'] = pp['VolumeInVoxels'] * spc
    # Ensure that the label column is of integer type - if there is any NaN, it will be float
    # Something has gone seriously wrong if the labels are not interpreted as integers
    if not np.issubdtype(pp['Label'].dtype, np.integer):
        raise ValueError('Label column not integer type, label values may be invalid')
    return pp






__all__ = ['iMath',
           'image_math',
           'multiply_images',
           'iMath_get_largest_component',
           'iMath_normalize',
           'iMath_truncate_intensity',
           'iMath_sharpen',
           'iMath_pad',
           'iMath_maurer_distance',
           'iMath_perona_malik',
           'iMath_grad',
           'iMath_laplacian',
           'iMath_canny',
           'iMath_histogram_equalization',
           'iMath_MD',
           'iMath_ME',
           'iMath_MO',
           'iMath_MC',
           'iMath_GD',
           'iMath_GE',
           'iMath_GO',
           'iMath_GC',
           'iMath_fill_holes',
           'iMath_get_largest_component',
           'iMath_normalize',
           'iMath_truncate_intensity',
           'iMath_sharpen',
           'iMath_propagate_labels_through_mask']

from ants.internal import get_lib_fn, process_arguments
from ants.decorators import image_method

_iMathOps = {'FillHoles',
            'GetLargestComponent',
            'Normalize',
            'Sharpen',
            'Pad',
            'D',
            'MaurerDistance',
            'PeronaMalik',
            'Grad',
            'Laplacian',
            'Canny',
            'HistogramEqualization',
            'MD',
            'ME',
            'MO',
            'MC',
            'GD',
            'GE',
            'GO',
            'GC',
            'FillHoles',
            'GetLargestComponent',
            'LabelStats',
            'Normalize',
            'TruncateIntensity',
            'Sharpen',
            'PropagateLabelsThroughMask'}

@image_method
def multiply_images(image1, image2):
    return image1 * image2


@image_method
def iMath(image, operation, *args):
    """
    Perform various (often mathematical) operations on the input image/s.
    Additional parameters should be specific for each operation.
    See the the full iMath in ANTs, on which this function is based.

    ANTsR function: `iMath`

    Arguments
    ---------
    image : ANTsImage
        input object, usually antsImage

    operation
        a string e.g. "GetLargestComponent" ... the special case of "GetOperations"
        or "GetOperationsFull" will return a list of operations and brief
        description. Some operations may not be valid (WIP), but most are.

    *args : non-keyword arguments
        additional parameters specific to the operation

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> img2 = ants.iMath(img, 'Canny', 1, 5, 12)
    """
    if operation not in _iMathOps:
        raise ValueError('Operation not recognized')

    imagedim = image.dimension
    outimage = image.clone()
    args = [imagedim, outimage, operation, image] + [a for a in args]
    processed_args = process_arguments(args)

    libfn = get_lib_fn('iMath')
    libfn(processed_args)
    return outimage
image_math = iMath


def iMath_ops():
    return _iMathOps

@image_method
def iMath_canny(image, sigma, lower, upper):
    return iMath(image, 'Canny', sigma, lower, upper)

@image_method
def iMath_fill_holes(image, hole_type=2):
    return iMath(image, 'FillHoles', hole_type)

@image_method
def iMath_GC(image, radius=1):
    return iMath(image, 'GC', radius)

@image_method
def iMath_GD(image, radius=1):
    return iMath(image, 'GD', radius)

@image_method
def iMath_GE(image, radius=1):
    return iMath(image, 'GE', radius)

@image_method
def iMath_GO(image, radius=1):
    return iMath(image, 'GO', radius)

@image_method
def iMath_get_largest_component(image, min_size=50):
    return iMath(image, 'GetLargestComponent', min_size)

@image_method
def iMath_grad(image, sigma=0.5, normalize=False):
    return iMath(image, 'Grad', sigma, normalize)

@image_method
def iMath_histogram_equalization(image, alpha, beta):
    return iMath(image, 'HistogramEqualization', alpha, beta)

@image_method
def iMath_laplacian(image, sigma=0.5, normalize=False):
    return iMath(image, 'Laplacian', sigma, normalize)

@image_method
def iMath_MC(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False):
    return iMath(image, 'MC', radius, value, shape, parametric, lines, thickness, include_center)

@image_method
def iMath_MD(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False):
    return iMath(image, 'MD', radius, value, shape, parametric, lines, thickness, include_center)

@image_method
def iMath_ME(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False):
    return iMath(image, 'ME', radius, value, shape, parametric, lines, thickness, include_center)

@image_method
def iMath_MO(image, radius=1, value=1, shape=1, parametric=False, lines=3, thickness=1, include_center=False):
    return iMath(image, 'MO', radius, value, shape, parametric, lines, thickness, include_center)

@image_method
def iMath_maurer_distance(image, foreground=1):
    return iMath(image, 'MaurerDistance', foreground)

@image_method
def iMath_normalize(image):
    return iMath(image, 'Normalize')

@image_method
def iMath_pad(image, padding):
    return iMath(image, 'Pad', padding)

@image_method
def iMath_perona_malik(image, conductance=0.25, n_iterations=1):
    return iMath(image, 'PeronaMalik', conductance, n_iterations)

@image_method
def iMath_sharpen(image):
    return iMath(image, 'Sharpen')

@image_method
def iMath_propagate_labels_through_mask(image, labels, stopping_value=100, propagation_method=0):
    """
    >>> import ants
    >>> wms = ants.image_read('~/desktop/wms.nii.gz')
    >>> thal = ants.image_read('~/desktop/thal.nii.gz')
    >>> img2 = ants.iMath_propagate_labels_through_mask(wms, thal, 500, 0)
    """
    return iMath(image, 'PropagateLabelsThroughMask', labels, stopping_value, propagation_method)

@image_method
def iMath_truncate_intensity(image, lower_q, upper_q, n_bins=64):
    """
    >>> import ants
    >>> img = ants.image_read(ants.get_ants_data('r16'))
    >>> ants.iMath_truncate_intensity( img, 0.2, 0.8 )
    """
    return iMath(image, 'TruncateIntensity', lower_q, upper_q, n_bins )



"""
Test ants_image.py

nptest.assert_allclose
self.assertEqual
self.assertTrue


"""

import os
import unittest
from common import run_tests
from tempfile import mktemp

import numpy as np

import numpy.testing as nptest

import ants

class TestModule_ants_image_io(unittest.TestCase):

    def setUp(self):
        img2d = ants.image_read(ants.get_ants_data('r16')).clone('float')
        img3d = ants.image_read(ants.get_ants_data('mni')).clone('float')
        arr2d = np.random.randn(69,70).astype('float32')
        arr3d = np.random.randn(69,70,71).astype('float32')
        vecimg2d = ants.from_numpy(np.random.randn(69,70,4), has_components=True)
        vecimg3d = ants.from_numpy(np.random.randn(69,70,71,2), has_components=True)
        self.imgs = [img2d, img3d]
        self.arrs = [arr2d, arr3d]
        self.vecimgs = [vecimg2d, vecimg3d]
        self.pixeltypes = ['unsigned char', 'unsigned int', 'float']

    def tearDown(self):
        pass

    def test_from_numpy(self):
        self.setUp()

        # no physical space info
        for arr in self.arrs:
            img = ants.from_numpy(arr)

            self.assertTrue(img.dimension, arr.ndim)
            self.assertTrue(img.shape, arr.shape)
            self.assertTrue(img.dtype, arr.dtype.name)
            nptest.assert_allclose(img.numpy(), arr)

            new_origin = tuple([6.9]*arr.ndim)
            new_spacing = tuple([3.6]*arr.ndim)
            new_direction = np.eye(arr.ndim)*9.6
            img2 = ants.from_numpy(arr, origin=new_origin, spacing=new_spacing, direction=new_direction)

            self.assertEqual(img2.origin, new_origin)
            self.assertEqual(img2.spacing, new_spacing)
            nptest.assert_allclose(img2.direction, new_direction)

        # test with components
        arr2d_components = np.random.randn(69,70,4).astype('float32')
        img = ants.from_numpy(arr2d_components, has_components=True)
        self.assertEqual(img.components, arr2d_components.shape[-1])
        nptest.assert_allclose(arr2d_components, img.numpy())

    def test_make_image(self):
        self.setUp()

        for arr in self.arrs:
            voxval = 6.
            img = ants.make_image(arr.shape, voxval=voxval)
            self.assertTrue(img.dimension, arr.ndim)
            self.assertTrue(img.shape, arr.shape)
            nptest.assert_allclose(img.mean(), voxval)

            new_origin = tuple([6.9]*arr.ndim)
            new_spacing = tuple([3.6]*arr.ndim)
            new_direction = np.eye(arr.ndim)*9.6
            img2 = ants.make_image(arr.shape, voxval=voxval, origin=new_origin, spacing=new_spacing, direction=new_direction)

            self.assertTrue(img2.dimension, arr.ndim)
            self.assertTrue(img2.shape, arr.shape)
            nptest.assert_allclose(img2.mean(), voxval)
            self.assertEqual(img2.origin, new_origin)
            self.assertEqual(img2.spacing, new_spacing)
            nptest.assert_allclose(img2.direction, new_direction)

            for ptype in self.pixeltypes:
                img = ants.make_image(arr.shape, voxval=1., pixeltype=ptype)
                self.assertEqual(img.pixeltype, ptype)

        # test with components
        img = ants.make_image((69,70,4), has_components=True)
        self.assertEqual(img.components, 4)
        self.assertEqual(img.dimension, 2)
        nptest.assert_allclose(img.mean(), 0.)

        img = ants.make_image((69,70,71,4), has_components=True)
        self.assertEqual(img.components, 4)
        self.assertEqual(img.dimension, 3)
        nptest.assert_allclose(img.mean(), 0.)

        # set from image
        for img in self.imgs:
            mask = ants.image_clone(  img > img.mean(), pixeltype = 'float' )
            arr = img[mask]
            img2 = ants.make_image(mask, voxval=arr)
            nptest.assert_allclose(img2.numpy(), (img*mask).numpy())
            self.assertTrue(ants.image_physical_space_consistency(img2,mask))

            # set with arr.ndim > 1
            img2 = ants.make_image(mask, voxval=np.expand_dims(arr,-1))
            nptest.assert_allclose(img2.numpy(), (img*mask).numpy())
            self.assertTrue(ants.image_physical_space_consistency(img2,mask))

            #with self.assertRaises(Exception):
            #    # wrong number of non-zero voxels
            #    img3 = ants.make_image(img, voxval=arr)


    def test_matrix_to_images(self):
        # def matrix_to_images(data_matrix, mask):
        for img in self.imgs:
            imgmask = ants.image_clone(  img > img.mean(), pixeltype = 'float' )
            data = img[imgmask]
            dataflat = data.reshape(1,-1)
            mat = np.vstack([dataflat,dataflat]).astype('float32')
            imglist = ants.matrix_to_images(mat, imgmask)
            nptest.assert_allclose((img*imgmask).numpy(), imglist[0].numpy())
            nptest.assert_allclose((img*imgmask).numpy(), imglist[1].numpy())
            self.assertTrue(ants.image_physical_space_consistency(img,imglist[0]))
            self.assertTrue(ants.image_physical_space_consistency(img,imglist[1]))

            # go back to matrix
            mat2 = ants.images_to_matrix(imglist, imgmask)
            nptest.assert_allclose(mat, mat2)

            # test with matrix.ndim > 2
            img = img.clone()
            img.set_direction(img.direction*2)
            imgmask = ants.image_clone(  img > img.mean(), pixeltype = 'float' )
            arr = (img*imgmask).numpy()
            arr = arr[arr>=0.5]
            arr2 = arr.copy()
            mat = np.stack([arr,arr2])
            imglist = ants.matrix_to_images(mat, imgmask)
            for im in imglist:
                self.assertTrue(ants.allclose(im, imgmask*img))
                self.assertTrue(ants.image_physical_space_consistency(im, imgmask))

            # test for wrong number of voxels
            #with self.assertRaises(Exception):
            #    arr = (img*imgmask).numpy()
            #    arr = arr[arr>0.5]
            #    arr2 = arr.copy()
            #    mat = np.stack([arr,arr2])
            #    imglist = ants.matrix_to_images(mat, img)

    def test_images_to_matrix(self):
        # def images_to_matrix(image_list, mask=None, sigma=None, epsilon=0):
        for img in self.imgs:
            mask = ants.image_clone(  img > img.mean(), pixeltype = 'float' )
            imglist = [img.clone(),img.clone(),img.clone()]
            imgmat = ants.images_to_matrix(imglist, mask=mask)
            self.assertTrue(imgmat.shape[0] == len(imglist))
            self.assertTrue(imgmat.shape[1] == (mask>0).sum())
            self.assertTrue(np.allclose(img[mask], imgmat[0,:]))

            # go back to images
            imglist2 = ants.matrix_to_images(imgmat, mask)
            for i1,i2 in zip(imglist,imglist2):
                self.assertTrue(ants.image_physical_space_consistency(i1,i2))
                nptest.assert_allclose(i1.numpy()*mask.numpy(),i2.numpy())

            if img.dimension == 2:
                # with sigma
                mask = ants.image_clone(  img > img.mean(), pixeltype = 'float' )
                imglist = [img.clone(),img.clone(),img.clone()]
                imgmat = ants.images_to_matrix(imglist, mask=mask, sigma=2.)

                # with no mask
                imgmat = ants.images_to_matrix(imglist)

                # Mask not binary
                mask = ants.image_clone(  img / img.mean(), pixeltype = 'float' )
                imgmat = ants.images_to_matrix(imglist, mask=mask, epsilon=1)

                # with mask of different shape
                s = [65]*img.dimension
                mask2 = ants.from_numpy(np.random.randn(*s), spacing=[4.0, 4.0])
                mask2 = mask2 > mask2.mean()
                imgmat = ants.images_to_matrix(imglist, mask=mask2)
                self.assertTrue(imgmat.shape[0] == len(imglist))
                self.assertTrue(imgmat.shape[1] == (mask2>0).sum())



    def timeseries_to_matrix(self):
        img = ants.make_image( (10,10,10,5 ) )
        mat = ants.timeseries_to_matrix( img )

        img = ants.make_image( (10,10,10,5 ) )
        mask = ants.ndimage_to_list( img )[0] * 0
        mask[ 4:8, 4:8, 4:8 ] = 1
        mat = ants.timeseries_to_matrix( img, mask = mask )
        img2 = ants.matrix_to_timeseries( img,  mat, mask)

    def test_image_header_info(self):
        # def image_header_info(filename):
        for img in self.imgs:
            img.set_spacing([6.9]*img.dimension)
            img.set_origin([3.6]*img.dimension)
            tmpfile = mktemp(suffix='.nii.gz')
            ants.image_write(img, tmpfile)

            info = ants.image_header_info(tmpfile)
            self.assertEqual(info['dimensions'], img.shape)
            nptest.assert_allclose(info['direction'], img.direction)
            self.assertEqual(info['nComponents'], img.components)
            self.assertEqual(info['nDimensions'], img.dimension)
            self.assertEqual(info['origin'], img.origin)
            self.assertEqual(info['pixeltype'], img.pixeltype)
            self.assertEqual(info['pixelclass'], 'vector' if img.has_components else 'scalar')
            self.assertEqual(info['spacing'], img.spacing)

            try:
                os.remove(tmpfile)
            except:
                pass

        # test on vector image
        img = ants.from_numpy(np.random.randn(69,60,4).astype('float32'), has_components=True)
        tmpfile = mktemp(suffix='.nii.gz')
        ants.image_write(img, tmpfile)
        info = ants.image_header_info(tmpfile)
        self.assertEqual(info['dimensions'], img.shape)
        nptest.assert_allclose(info['direction'], img.direction)
        self.assertEqual(info['nComponents'], img.components)
        self.assertEqual(info['nDimensions'], img.dimension)
        self.assertEqual(info['origin'], img.origin)
        self.assertEqual(info['pixeltype'], img.pixeltype)
        self.assertEqual(info['pixelclass'], 'vector' if img.has_components else 'scalar')
        self.assertEqual(info['spacing'], img.spacing)


        img = ants.from_numpy(np.random.randn(69,60,70,2).astype('float32'), has_components=True)
        tmpfile = mktemp(suffix='.nii.gz')
        ants.image_write(img, tmpfile)
        info = ants.image_header_info(tmpfile)
        self.assertEqual(info['dimensions'], img.shape)
        nptest.assert_allclose(info['direction'], img.direction)
        self.assertEqual(info['nComponents'], img.components)
        self.assertEqual(info['nDimensions'], img.dimension)
        self.assertEqual(info['origin'], img.origin)
        self.assertEqual(info['pixeltype'], img.pixeltype)
        self.assertEqual(info['pixelclass'], 'vector' if img.has_components else 'scalar')
        self.assertEqual(info['spacing'], img.spacing)

        # non-existant file
        with self.assertRaises(Exception):
            tmpfile = mktemp(suffix='.nii.gz')
            ants.image_header_info(tmpfile)

    def test_image_clone(self):
        for img in self.imgs:
            img = ants.image_clone(img, 'unsigned char')
            orig_ptype = img.pixeltype
            for ptype in self.pixeltypes:
                imgcloned = ants.image_clone(img, ptype)
                self.assertTrue(ants.image_physical_space_consistency(img,imgcloned))
                nptest.assert_allclose(img.numpy(), imgcloned.numpy())
                self.assertEqual(imgcloned.pixeltype, ptype)
                self.assertEqual(img.pixeltype, orig_ptype)

        for img in self.vecimgs:
            img = img.clone('unsigned char')
            orig_ptype = img.pixeltype
            for ptype in self.pixeltypes:
                imgcloned = ants.image_clone(img, ptype)
                self.assertTrue(ants.image_physical_space_consistency(img,imgcloned))
                self.assertEqual(imgcloned.components, img.components)
                nptest.assert_allclose(img.numpy(), imgcloned.numpy())
                self.assertEqual(imgcloned.pixeltype, ptype)
                self.assertEqual(img.pixeltype, orig_ptype)

    def test_image_read_write(self):
        # def image_read(filename, dimension=None, pixeltype='float'):
        # def image_write(image, filename):

        # test scalar images
        for img in self.imgs:
            img = (img - img.min()) / (img.max() - img.min())
            img = img * 255.
            img = img.clone('unsigned char')
            for ptype in self.pixeltypes:
                img = img.clone(ptype)
                tmpfile = mktemp(suffix='.nii.gz')
                ants.image_write(img, tmpfile)

                img2 = ants.image_read(tmpfile)
                self.assertTrue(ants.image_physical_space_consistency(img,img2))
                self.assertEqual(img2.components, img.components)
                nptest.assert_allclose(img.numpy(), img2.numpy())

            # unsupported ptype
            with self.assertRaises(Exception):
                ants.image_read(tmpfile, pixeltype='not-suppoted-ptype')

        # test vector images
        for img in self.vecimgs:
            img = (img - img.min()) / (img.max() - img.min())
            img = img * 255.
            img = img.clone('unsigned char')
            for ptype in self.pixeltypes:
                img = img.clone(ptype)
                tmpfile = mktemp(suffix='.nii.gz')
                ants.image_write(img, tmpfile)

                img2 = ants.image_read(tmpfile)
                self.assertTrue(ants.image_physical_space_consistency(img,img2))
                self.assertEqual(img2.components, img.components)
                nptest.assert_allclose(img.numpy(), img2.numpy())

        # test saving/loading as npy
        for img in self.imgs:
            tmpfile = mktemp(suffix='.npy')
            ants.image_write(img, tmpfile)
            img2 = ants.image_read(tmpfile)

            self.assertTrue(ants.image_physical_space_consistency(img,img2))
            self.assertEqual(img2.components, img.components)
            nptest.assert_allclose(img.numpy(), img2.numpy())

            # with no json header
            arr = img.numpy()
            tmpfile = mktemp(suffix='.npy')
            np.save(tmpfile, arr)
            img2 = ants.image_read(tmpfile)
            nptest.assert_allclose(img.numpy(), img2.numpy())

        # non-existant file
        with self.assertRaises(ValueError):
            tmpfile = mktemp(suffix='.nii.gz')
            ants.image_read(tmpfile)

        # Test empty file
        with self.assertRaises(RuntimeError):
            tmpfile = mktemp(suffix='.nii.gz')
            open(tmpfile, 'a').close()
            ants.image_read(tmpfile)


    def test_image_from_numpy_shape(self):
        arr = np.random.randn(1,9,1).astype('float32')
        img = ants.from_numpy(arr)
        self.assertEqual(img.shape, arr.shape)
        nptest.assert_allclose(img.numpy(), arr)

        arr = np.random.randn(2,3,4).astype('float32')
        img = ants.from_numpy(arr)
        self.assertEqual(img.shape, arr.shape)
        nptest.assert_allclose(img.numpy(), arr)

        # Test special case where shape is (1,N), was getting transposed
        arr = np.random.randn(1,9).astype('float32')
        img = ants.from_numpy(arr)
        self.assertEqual(img.shape, arr.shape)
        nptest.assert_allclose(img.numpy(), arr)

        arr = np.random.randn(9,1).astype('float32')
        img = ants.from_numpy(arr)
        self.assertEqual(img.shape, arr.shape)
        nptest.assert_allclose(img.numpy(), arr)

if __name__ == '__main__':
    run_tests()




from tempfile import mktemp
import os

import ants
from ants.internal import get_lib_fn, process_arguments

__all__ = ['average_affine_transform', 
           'average_affine_transform_no_rigid']


def _average_affine_transform_driver(transformlist, referencetransform=None, funcname="AverageAffineTransform"):
    """
    takes a list of transforms (files at the moment)
    and returns the average
    """

    # AverageAffineTransform deals with transform files,
    # so this function will need to deal with already
    # loaded files. Doesn't look like the magic
    # available for images has been added for transforms.
    res_temp_file = mktemp(suffix='.mat')

    # could do some stuff here to cope with transform lists that
    # aren't files
    
    # load one of the transforms to figure out the dimension
    tf = ants.read_transform(transformlist[0])
    if referencetransform is None:
        args = [tf.dimension, res_temp_file] + transformlist
    else:
        args = [tf.dimension, res_temp_file] + ['-R',  referencetransform] + transformlist
    pargs = process_arguments(args)
    print(pargs)
    libfun = get_lib_fn(funcname)
    status = libfun(pargs)

    res = ants.read_transform(res_temp_file)
    os.remove(res_temp_file)
    return res

def average_affine_transform(transformlist, referencetransform=None):
    return _average_affine_transform_driver(transformlist, referencetransform, "AverageAffineTransform")


def average_affine_transform_no_rigid(transformlist, referencetransform=None):
    return _average_affine_transform_driver(transformlist, referencetransform, "AverageAffineTransformNoRigid")
    




__all__ = ['pad_image']

import math

import ants
from ants.decorators import image_method
from ants.internal import get_lib_fn

@image_method
def pad_image(image, shape=None, pad_width=None, value=0.0, return_padvals=False):
    """
    Pad an image to have the given shape or to be isotropic.

    Arguments
    ---------
    image : ANTsImage
        image to pad

    shape : tuple
        - if shape is given, the image will be padded in each dimension
          until it has this shape
        - if shape and pad_width are both None, the image will be padded along
          each dimension to match the largest existing dimension so that it has
          isotropic dimensions.

    pad_width : list of integers or list-of-list of integers
        How much to pad in each direction. If a single list is
        supplied (e.g., [4,4,4]), then the image will be padded by half
        that amount on both sides. If a list-of-list is supplied
        (e.g., [(0,4),(0,4),(0,4)]), then the image will be
        padded unevenly on the different sides

    pad_value : scalar
        value with which image will be padded

    Example
    -------
    >>> import ants
    >>> img = ants.image_read(ants.get_data('r16'))
    >>> img2 = ants.pad_image(img, shape=(300,300))
    >>> mni = ants.image_read(ants.get_data('mni'))
    >>> mni2 = ants.pad_image(mni)
    >>> mni3 = ants.pad_image(mni, pad_width=[(0,4),(0,4),(0,4)])
    >>> mni4 = ants.pad_image(mni, pad_width=(4,4,4))
    """
    inpixeltype = image.pixeltype
    ndim = image.dimension
    if image.pixeltype != 'float':
        image = image.clone('float')

    if pad_width is None:
        if shape is None:
            shape = [max(image.shape)] * image.dimension
        lower_pad_vals = [math.floor(max(ns-os,0)/2) for os,ns in zip(image.shape, shape)]
        upper_pad_vals = [math.ceil(max(ns-os,0)/2) for os,ns in zip(image.shape, shape)]
    else:
        if shape is not None:
            raise ValueError('Cannot give both `shape` and `pad_width`. Pick one!')
        if len(pad_width) != image.dimension:
            raise ValueError('Must give pad width for each image dimension')

        lower_pad_vals = []
        upper_pad_vals = []
        for p in pad_width:
            if isinstance(p, (list, tuple)):
                lower_pad_vals.append(p[0])
                upper_pad_vals.append(p[1])
            else:
                lower_pad_vals.append(math.floor(p/2))
                upper_pad_vals.append(math.ceil(p/2))

    libfn = get_lib_fn('padImage')
    itkimage = libfn(image.pointer, lower_pad_vals, upper_pad_vals, value)

    new_image = ants.from_pointer(itkimage).clone(inpixeltype)
    if return_padvals:
        return new_image, lower_pad_vals, upper_pad_vals
    else:
        return new_image


